from __future__ import division

import os

import matplotlib.pyplot as plt
import pandas as pd

from pyomo.environ import *


def build_model():
    m = ConcreteModel()
    m.months = RangeSet(0, 120, doc="10 years")
    m.discount_rate = Param(initialize=0.08, doc="8%")
    m.conv_setup_time = Param(initialize=12)
    m.modular_setup_time = Param(initialize=3)
    m.modular_move_time = Param(initialize=3)
    m.markets = RangeSet(5)

    xls_data = pd.read_excel(
        os.path.join(os.path.dirname(__file__), "multiple_market_size.xlsx"),
        sheet_name=["demand", "locations"], index_col=0)

    @m.Param(m.markets, m.months)
    def market_demand(m, mkt, mo):
        return float(xls_data["demand"]["market%s" % mkt][mo])

    @m.Param(m.months, doc="Cost of transporting unit of product one mile.")
    def transport_cost(m, mo):
        return 0.2 / 12 * (1 + m.discount_rate / 12) ** (-mo / 12)

    m.route_fixed_cost = Param(
        initialize=100,
        doc="Cost of establishing a route from a modular site to a market")

    m.conv_x = Var(bounds=(0, 300), doc="x-coordinate of centralized plant.")
    m.conv_y = Var(bounds=(0, 300), doc="y-coordinate of centralized plant.")

    m.conv_size = Var(bounds=(10, 500), initialize=10,
                      doc="Size of conventional plant.")
    m.conv_cost = Var()
    m.conv_base_cost = Param(initialize=1000, doc="Cost for size 20")
    m.conv_exponent = Param(initialize=0.6)

    m.cost_calc = Constraint(
        expr=m.conv_cost == (
            m.conv_base_cost * (m.conv_size / 20) ** m.conv_exponent))

    @m.Param(m.markets)
    def mkt_x(m, mkt):
        return float(xls_data["locations"]["x"]["market%s" % mkt])

    @m.Param(m.markets)
    def mkt_y(m, mkt):
        return float(xls_data["locations"]["y"]["market%s" % mkt])

    m.dist_to_mkt = Var(m.markets, bounds=(0, sqrt(300**2 + 300**2)))

    @m.Constraint(m.markets)
    def distance_calculation(m, mkt):
        return m.dist_to_mkt[mkt] == sqrt(
            (m.conv_x / 300 - m.mkt_x[mkt] / 300)**2 +
            (m.conv_y / 300 - m.mkt_y[mkt] / 300)**2) * 300

    m.shipments_to_mkt = Var(m.markets, m.months, bounds=(0, 100))
    m.production = Var(m.months, bounds=(0, 500))

    @m.Constraint(m.months)
    def production_satisfaction(m, mo):
        return m.production[mo] == sum(m.shipments_to_mkt[mkt, mo]
                                       for mkt in m.markets)

    @m.Constraint(m.months)
    def size_requirement(m, mo):
        if mo < m.conv_setup_time:
            return m.production[mo] == 0
        else:
            return m.production[mo] <= m.conv_size

    @m.Constraint(m.markets, m.months)
    def demand_satisfaction(m, mkt, mo):
        return m.market_demand[mkt, mo] <= m.shipments_to_mkt[mkt, mo]

    m.variable_shipment_cost = Expression(
        expr=sum(
            m.shipments_to_mkt[mkt, mo] * m.dist_to_mkt[mkt] *
            m.transport_cost[mo]
            for mkt in m.markets for mo in m.months))

    m.total_cost = Objective(
        expr=m.variable_shipment_cost + m.conv_cost + 5 * m.route_fixed_cost)

    return m


def build_modular_model():
    m = ConcreteModel()
    m.bigM = Suffix(direction=Suffix.LOCAL)
    m.months = RangeSet(0, 120, doc="10 years")
    m.discount_rate = Param(initialize=0.08, doc="8%")
    m.modular_setup_time = Param(initialize=3)
    m.modular_move_time = Param(initialize=3)
    m.markets = RangeSet(5)
    m.modular_sites = RangeSet(2)
    m.site_pairs = Set(
        initialize=m.modular_sites * m.modular_sites,
        filter=lambda _, x, y: not x == y)
    m.unique_site_pairs = Set(
        initialize=m.site_pairs, filter=lambda _, x, y: x < y)

    xls_data = pd.read_excel(
        os.path.join(os.path.dirname(__file__), "multiple_market_size.xlsx"),
        sheet_name=["demand", "locations"], index_col=0)

    @m.Param(m.markets, m.months)
    def market_demand(m, mkt, mo):
        return float(xls_data["demand"]["market%s" % mkt][mo])

    @m.Param(m.months, doc="Cost of transporting unit of product one mile.")
    def transport_cost(m, mo):
        return 0.2 / 12 * (1 + m.discount_rate / 12) ** (-mo / 12)

    m.route_fixed_cost = Param(
        initialize=100,
        doc="Cost of establishing a route from a modular site to a market")

    @m.Param(m.months, doc="Cost of transporting a module one mile")
    def modular_transport_cost(m, mo):
        return 1 * (1 + m.discount_rate / 12) ** (-mo / 12)
    m.modular_base_cost = Param(initialize=1000, doc="Cost for size 20")

    @m.Param(m.months)
    def modular_unit_cost(m, mo):
        return m.modular_base_cost * (1 + m.discount_rate / 12) ** (-mo / 12)

    @m.Param(m.markets)
    def mkt_x(m, mkt):
        return float(xls_data["locations"]["x"]["market%s" % mkt])

    @m.Param(m.markets)
    def mkt_y(m, mkt):
        return float(xls_data["locations"]["y"]["market%s" % mkt])

    m.site_x = Var(
        m.modular_sites, bounds=(0, 300), initialize={
            1: 50, 2: 225, 3: 250})
    m.site_y = Var(
        m.modular_sites, bounds=(0, 300), initialize={
            1: 300, 2: 275, 3: 50})

    m.num_modules = Var(
        m.modular_sites, m.months, domain=Integers, bounds=(0, 25))
    m.modules_transferred = Var(
        m.site_pairs, m.months,
        domain=Integers, bounds=(0, 25),
        doc="Number of modules moved from one site to another in a month.")
    # m.modules_transferred[...].fix(0)
    m.modules_added = Var(
        m.modular_sites, m.months, domain=Integers, bounds=(0, 25))

    m.dist_to_mkt = Var(
        m.modular_sites, m.markets, bounds=(0, sqrt(300**2 + 300**2)))
    m.sqr_scaled_dist_to_mkt = Var(
        m.modular_sites, m.markets, bounds=(0.001, 8), initialize=0.001)
    m.dist_to_site = Var(m.site_pairs, bounds=(0, sqrt(300**2 + 300**2)))
    m.sqr_scaled_dist_to_site = Var(
        m.site_pairs, bounds=(0.001, 8), initialize=0.001)
    m.shipments_to_mkt = Var(
        m.modular_sites, m.markets, m.months, bounds=(0, 100))
    m.production = Var(m.modular_sites, m.months, bounds=(0, 500))

    @m.Disjunct(m.modular_sites)
    def site_active(disj, site):
        @disj.Constraint(m.months)
        def production_satisfaction(site_disj, mo):
            return m.production[site, mo] == sum(
                m.shipments_to_mkt[site, mkt, mo] for mkt in m.markets)

        @disj.Constraint(m.months)
        def production_limit(site_disj, mo):
            return m.production[site, mo] <= m.num_modules[site, mo] * 20

        @disj.Constraint(m.months)
        def module_balance(site_disj, mo):
            existing_modules = m.num_modules[site, mo - 1] if mo >= 1 else 0
            new_modules = (m.modules_added[site, mo - m.modular_setup_time]
                           if mo > m.modular_setup_time else 0)
            xfrd_in_modules = sum(
                m.modules_transferred[from_site,
                                      site, mo - m.modular_move_time]
                for from_site in m.modular_sites
                if (not from_site == site) and mo > m.modular_move_time)
            xfrd_out_modules = sum(
                m.modules_transferred[site, to_site, mo]
                for to_site in m.modular_sites if not to_site == site)
            return m.num_modules[site, mo] == (
                new_modules + xfrd_in_modules - xfrd_out_modules +
                existing_modules)

    @m.Disjunct(m.modular_sites)
    def site_inactive(disj, site):
        disj.no_modules = Constraint(
            expr=sum(m.num_modules[site, mo] for mo in m.months) == 0)

        disj.no_module_transfer = Constraint(
            expr=sum(m.modules_transferred[site1, site2, mo]
                     for site1, site2 in m.site_pairs
                     for mo in m.months
                     if site1 == site or site2 == site) == 0)

        disj.no_shipments = Constraint(
            expr=sum(m.shipments_to_mkt[site, mkt, mo]
                     for mkt in m.markets for mo in m.months) == 0)

    @m.Disjunction(m.modular_sites)
    def site_active_or_not(m, site):
        return [m.site_active[site], m.site_inactive[site]]

    @m.Constraint(m.modular_sites, doc="Symmetry breaking for site activation")
    def site_active_ordering(m, site):
        if site + 1 <= max(m.modular_sites):
            return (m.site_active[site].indicator_var >=
                    m.site_active[site + 1].indicator_var)
        else:
            return Constraint.NoConstraint

    @m.Disjunct(m.unique_site_pairs)
    def pair_active(disj, site1, site2):
        disj.site1_active = Constraint(
            expr=m.site_active[site1].indicator_var == 1)
        disj.site2_active = Constraint(
            expr=m.site_active[site2].indicator_var == 1)
        disj.site_distance_calc = Constraint(
            expr=m.dist_to_site[site1, site2] == sqrt(
                m.sqr_scaled_dist_to_site[site1, site2]) * 150)
        disj.site_distance_symmetry = Constraint(
            expr=m.dist_to_site[site1, site2] == m.dist_to_site[site2, site1])
        disj.site_sqr_distance_calc = Constraint(
            expr=m.sqr_scaled_dist_to_site[site1, site2] == (
                (m.site_x[site1] / 150 - m.site_x[site2] / 150)**2 +
                (m.site_y[site1] / 150 - m.site_y[site2] / 150)**2))
        disj.site_sqr_distance_symmetry = Constraint(
            expr=m.sqr_scaled_dist_to_site[site1, site2] ==
            m.sqr_scaled_dist_to_site[site2, site1])

    @m.Disjunct(m.unique_site_pairs)
    def pair_inactive(disj, site1, site2):
        disj.site1_inactive = Constraint(
            expr=m.site_active[site1].indicator_var == 0)
        disj.site2_inactive = Constraint(
            expr=m.site_active[site2].indicator_var == 0)

        disj.no_module_transfer = Constraint(
            expr=sum(m.modules_transferred[site1, site2, mo]
                     for mo in m.months) == 0)

    @m.Disjunction(m.unique_site_pairs)
    def site_pair_active_or_not(m, site1, site2):
        return [m.pair_active[site1, site2], m.pair_inactive[site1, site2]]

    @m.Constraint(m.markets, m.months)
    def demand_satisfaction(m, mkt, mo):
        return m.market_demand[mkt, mo] <= sum(
            m.shipments_to_mkt[site, mkt, mo] for site in m.modular_sites)

    @m.Disjunct(m.modular_sites, m.markets)
    def product_route_active(disj, site, mkt):
        disj.site_active = Constraint(
            expr=m.site_active[site].indicator_var == 1)

        @disj.Constraint()
        def market_distance_calculation(site_disj):
            return m.dist_to_mkt[site, mkt] == sqrt(
                m.sqr_scaled_dist_to_mkt[site, mkt]) * 150

        @disj.Constraint()
        def market_sqr_distance_calc(site_disj):
            return m.sqr_scaled_dist_to_mkt[site, mkt] == (
                (m.site_x[site] / 150 - m.mkt_x[mkt] / 150)**2 +
                (m.site_y[site] / 150 - m.mkt_y[mkt] / 150)**2)

    @m.Disjunct(m.modular_sites, m.markets)
    def product_route_inactive(disj, site, mkt):
        disj.no_shipments = Constraint(
            expr=sum(m.shipments_to_mkt[site, mkt, mo]
                     for mo in m.months) == 0)

    @m.Disjunction(m.modular_sites, m.markets)
    def product_route_active_or_not(m, site, mkt):
        return [m.product_route_active[site, mkt],
                m.product_route_inactive[site, mkt]]

    m.variable_shipment_cost = Expression(
        expr=sum(
            m.shipments_to_mkt[site, mkt, mo] * m.dist_to_mkt[site, mkt] *
            m.transport_cost[mo]
            for site in m.modular_sites for mkt in m.markets
            for mo in m.months))

    m.fixed_shipment_cost = Expression(
        expr=sum(m.product_route_active[site, mkt].indicator_var *
                 m.route_fixed_cost
                 for site in m.modular_sites for mkt in m.markets))

    m.module_purchase_cost = Expression(
        expr=sum(m.modules_added[site, mo] * m.modular_unit_cost[mo]
                 for site in m.modular_sites for mo in m.months))

    m.module_transfer_cost = Expression(expr=sum(
        m.modules_transferred[site1, site2, mo]
        * m.dist_to_site[site1, site2] * m.modular_transport_cost[mo]
        for site1, site2 in m.site_pairs for mo in m.months))

    m.total_cost = Objective(
        expr=m.variable_shipment_cost
        + m.fixed_shipment_cost
        + m.module_purchase_cost
        + m.module_transfer_cost)

    return m


if __name__ == "__main__":
    # m = build_model()
    # SolverFactory('ipopt').solve(m)
    # print("Plant at ({conv_x:3.0f}, {conv_y:3.0f}) size {conv_size:3.0f} "
    #       "cost {conv_cost:4.0f}.".format(
    #           conv_x=m.conv_x.value,
    #           conv_y=m.conv_y.value,
    #           conv_size=m.conv_size.value,
    #           conv_cost=m.conv_cost.value))
    # print("Total cost: %s" % value(m.total_cost.expr))
    # print("  Variable ship cost: %s" % value(m.variable_shipment_cost))
    # print("  Fix ship cost: %s" % value(500))
    # print("  Conv buy cost: %s" % value(m.conv_cost))
    # exit()

    m = build_modular_model()
    # SolverFactory('gdpopt').solve(
    #     m, tee=True, mip_solver="gams", nlp_solver="ipopt",
    #     mip_solver_args={
    #                      'io_options': {'add_options': [
    #                          'option reslim=30;']}})
    TransformationFactory('gdp.bigm').apply_to(m, bigM=10000)
    # TransformationFactory('gdp.chull').apply_to(m)
    res = SolverFactory('gams').solve(m, tee=True, io_options={
        'add_options': ['option reslim = 300;']})
    # SolverFactory('gams').solve(
    #     m, tee=True,
    #     io_options={
    #         'solver': 'baron',
    #         'add_options': ['option solver=baron;',
    #                         'option reslim = 600;']})
    # from pyomo.util.infeasible import log_infeasible_constraints
    # log_infeasible_constraints(m)

    def record_generator():
        for mo in m.months:
            yield (
                (mo,)
                + tuple(m.num_modules[site, mo].value
                        for site in m.modular_sites)
                + tuple(m.production[site, mo].value
                        for site in m.modular_sites)
                + tuple(m.shipments_to_mkt[site, mkt, mo].value
                        for site in m.modular_sites
                        for mkt in m.markets)
            )

    df = pd.DataFrame.from_records(
        record_generator(),
        columns=("Month",)
        + tuple("Num Site%s" % site for site in m.modular_sites)
        + tuple("Prod Site%s" % site for site in m.modular_sites)
        + tuple("Ship Site%s to Mkt%s" % (site, mkt)
                for site in m.modular_sites for mkt in m.markets))
    df.to_excel("multiple_modular_config.xlsx")
    print("Total cost: %s" % value(m.total_cost.expr))
    print("  Variable ship cost: %s" % value(m.variable_shipment_cost))
    print("  Fix ship cost: %s" % value(m.fixed_shipment_cost))
    print("  Module buy cost: %s" % value(m.module_purchase_cost))
    print("  Module xfer cost: %s" % value(m.module_transfer_cost))
    for site in m.modular_sites:
        print("Site {:1.0f} at ({:3.0f}, {:3.0f})".format(
            site, m.site_x[site].value, m.site_y[site].value))
        print("  Supplies markets {}".format(tuple(
            mkt for mkt in m.markets
            if m.product_route_active[site, mkt].indicator_var.value == 1)))


    if res.solver.termination_condition is not TerminationCondition.optimal:
        exit()

    plt.plot([x.value for x in m.site_x.values()],
             [y.value for y in m.site_y.values()], 'k.', markersize=12)
    plt.plot([x for x in m.mkt_x.values()],
             [y for y in m.mkt_y.values()], 'bo', markersize=12)
    for mkt in m.markets:
        plt.annotate('mkt%s' % mkt, (m.mkt_x[mkt], m.mkt_y[mkt]),
                     (m.mkt_x[mkt] + 2, m.mkt_y[mkt] + 0))
    for site in m.modular_sites:
        plt.annotate(
            'site%s' % site, (m.site_x[site].value, m.site_y[site].value),
             (m.site_x[site].value + 2, m.site_y[site].value + 0))
    for site, mkt in m.modular_sites * m.markets:
        if m.product_route_active[site, mkt].indicator_var.value == 1:
            plt.arrow(m.site_x[site].value, m.site_y[site].value,
                      m.mkt_x[mkt] - m.site_x[site].value,
                      m.mkt_y[mkt] - m.site_y[site].value,
                      width=0.8, length_includes_head=True, color='r')
    for site1, site2 in m.site_pairs:
        if sum(m.modules_transferred[site1, site2, mo].value
               for mo in m.months) > 0:
            plt.arrow(m.site_x[site1].value, m.site_y[site1].value,
                      m.site_x[site2].value - m.site_x[site1].value,
                      m.site_y[site2].value - m.site_y[site1].value,
                      width=0.9, length_includes_head=True,
                      linestyle='dotted', color='k')
    plt.show()



"""Heat integration case study.

This is example 1 of the Yee & Grossmann, 1990 paper "Simultaneous optimization
models for heat integration--II".
DOI: 10.1016/0098-1354(90)85010-8

This is a modification to support the incorporation of standardized exchanger
modules using integer variables for module selection.

"""
from __future__ import division

from pyomo.environ import (Constraint, Integers, log, Var)
from pyomo.gdp import Disjunct, Disjunction

from gdplib.mod_hens import common


def build_single_module(cafaro_approx, num_stages):
    m = build_model(cafaro_approx, num_stages)
    # Require modular
    for hot, cold, stg in m.valid_matches * m.stages:
        disj = m.exchanger_exists[hot, cold, stg]
        disj.modular.indicator_var.fix(1)
        disj.conventional.deactivate()

    # Must choose only one type of module
    @m.Disjunct(m.module_sizes)
    def module_type(disj, size):
        """Disjunct for selection of one module type."""
        disj.no_other_module_types = Constraint(
            expr=sum(m.num_modules[hot, cold, stage, area]
                     for hot, cold in m.valid_matches
                     for stage in m.stages
                     for area in m.module_sizes
                     if area != size) == 0)
    m.select_one_module_type = Disjunction(
        expr=[m.module_type[area] for area in m.module_sizes])

    return m


def build_require_modular(cafaro_approx, num_stages):
    m = build_model(cafaro_approx, num_stages)
    # Require modular
    for hot, cold, stg in m.valid_matches * m.stages:
        m.exchanger_exists[hot, cold, stg].conventional.deactivate()

    return m


def build_modular_option(cafaro_approx, num_stages):
    return build_model(cafaro_approx, num_stages)


def build_model(use_cafaro_approximation, num_stages):
    """Build the model."""
    m = common.build_model(use_cafaro_approximation, num_stages)

    m.num_modules = Var(
        m.valid_matches, m.stages, m.module_sizes,
        doc="The number of modules of each size at each exchanger.",
        domain=Integers, bounds=(0, 100), initialize=0)
    # improve quality of bounds
    for size in m.module_sizes:
        for var in m.num_modules[:, :, :, size]:
            var.setub(m.max_num_modules[size])

    for hot, cold, stg in m.valid_matches * m.stages:
        disj = m.exchanger_exists[hot, cold, stg]

        disj.conventional = Disjunct()
        if not use_cafaro_approximation:
            disj.conventional.exchanger_area_cost = Constraint(
                expr=m.exchanger_area_cost[stg, hot, cold] * 1E-3 >=
                m.exchanger_area_cost_factor[hot, cold] * 1E-3 *
                m.exchanger_area[stg, hot, cold] ** m.area_cost_exponent)
        else:
            disj.conventional.exchanger_area_cost = Constraint(
                expr=m.exchanger_area_cost[stg, hot, cold] * 1E-3 >=
                m.exchanger_area_cost_factor[hot, cold] * 1E-3 * m.cafaro_k
                * log(m.cafaro_b * m.exchanger_area[stg, hot, cold] + 1)
            )
        m.BigM[disj.conventional.exchanger_area_cost] = 100

        disj.conventional.exchanger_fixed_cost = Constraint(
            expr=m.exchanger_fixed_cost[stg, hot, cold] ==
            m.exchanger_fixed_unit_cost[hot, cold])

        @disj.conventional.Constraint(m.module_sizes)
        def no_modules(_, area):
            return m.num_modules[hot, cold, stg, area] == 0

        disj.modular = Disjunct()
        disj.modular.exchanger_area_cost = Constraint(
            expr=m.exchanger_area_cost[stg, hot, cold] * 1E-3 ==
            sum(m.module_area_cost[hot, cold, area]
                * m.num_modules[hot, cold, stg, area]
                for area in m.module_sizes)
            * 1E-3)
        disj.modular.exchanger_fixed_cost = Constraint(
            expr=m.exchanger_fixed_cost[stg, hot, cold] ==
            m.module_fixed_unit_cost * sum(m.num_modules[hot, cold, stg, area]
                                           for area in m.module_sizes))
        disj.modular.exchanger_area = Constraint(
            expr=m.exchanger_area[stg, hot, cold] ==
            sum(area * m.num_modules[hot, cold, stg, area]
                for area in m.module_sizes))
        disj.modular_or_not = Disjunction(
            expr=[disj.modular, disj.conventional])

        # Area requirement
        disj.exchanger_required_area = Constraint(
            expr=m.exchanger_area[stg, hot, cold] * (
                m.U[hot, cold] * m.LMTD[hot, cold, stg]) >=
            m.heat_exchanged[hot, cold, stg])
        m.BigM[disj.exchanger_required_area] = 5000

    return m



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

from pyomo.environ import *
from pyomo.gdp import *
from pyomo.common.fileutils import this_file_dir
from os.path import join

#
# Jobshop example from http://www.gams.com/modlib/libhtml/logmip4.htm
#
# This model solves a jobshop scheduling, which has a set of jobs
# which must be processed in sequence of stages but not all jobs
# require all stages. A zero wait transfer policy is assumed between
# stages. To obtain a feasible solution it is necessary to eliminate
# all clashes between jobs. It requires that no two jobs be performed
# at any stage at any time. The objective is to minimize the makespan,
# the time to complete all jobs.
#
# References:
#
# Raman & Grossmann, Computers and Chemical Engineering 18, 7, p.563-578, 1994.
#
# Aldo Vecchietti, LogMIP User's Manual, http://www.logmip.ceride.gov.ar/, 2007
#

def build_model():
    model = AbstractModel()

    model.JOBS = Set(ordered=True)
    model.STAGES = Set(ordered=True)
    model.I_BEFORE_K = RangeSet(0,1)

    # Task durations
    model.tau = Param(model.JOBS, model.STAGES, default=0)

    # Total Makespan (this will be the objective)
    model.ms = Var()
    # Start time of each job
    def t_bounds(model, I):
        return (0, sum(value(model.tau[idx]) for idx in model.tau))
    model.t = Var( model.JOBS, within=NonNegativeReals, bounds=t_bounds )

    # Auto-generate the L set (potential collisions between 2 jobs at any stage.
    def _L_filter(model, I, K, J):
        return I < K and model.tau[I,J] and model.tau[K,J]
    model.L = Set( initialize=model.JOBS * model.JOBS * model.STAGES,
                   dimen=3, filter=_L_filter)

    # Makespan is greater than the start time of every job + that job's
    # total duration
    def _feas(model, I):
        return model.ms >= model.t[I] + sum(model.tau[I,M] for M in model.STAGES)
    model.Feas = Constraint(model.JOBS, rule=_feas)

    # Disjunctions to prevent clashes at a stage: This creates a set of
    # disjunct pairs: one if job I occurs before job K and the other if job
    # K occurs before job I.
    def _NoClash(disjunct, I, K, J, IthenK):
        model = disjunct.model()
        lhs = model.t[I] + sum([M<J and model.tau[I,M] or 0 for M in model.STAGES])
        rhs = model.t[K] + sum([M<J and model.tau[K,M] or 0 for M in model.STAGES])
        if IthenK:
            disjunct.c = Constraint(expr=lhs+model.tau[I,J]<=rhs)
        else:
            disjunct.c = Constraint(expr=rhs+model.tau[K,J]<=lhs)
    model.NoClash = Disjunct(model.L, model.I_BEFORE_K, rule=_NoClash)

    # Define the disjunctions: either job I occurs before K or K before I
    def _disj(model, I, K, J):
        return [model.NoClash[I,K,J,IthenK] for IthenK in model.I_BEFORE_K]
    model.disj = Disjunction(model.L, rule=_disj)

    # minimize makespan
    model.makespan = Objective(expr=model.ms)
    return model


def build_small_concrete():
    return build_model().create_instance(join(this_file_dir(), 'jobshop-small.dat'))


if __name__ == "__main__":
    m = build_small_concrete()
    TransformationFactory('gdp.bigm').apply_to(m)
    SolverFactory('gams').solve(m, solver='baron', tee=True, add_options=['option optcr=1e-6;'])
    m.makespan.display()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pandas as pd
from pyomo.core.expr.logical_expr import *
from pyomo.core.plugins.transform.logical_to_linear import update_boolean_vars_from_binary
from pyomo.environ import *
from pyomo.environ import TerminationCondition as tc


def build_model():
    m = ConcreteModel()

    m.syngas_techs = Set(
        doc="syngas process technologies",
        initialize=['SMR', 'POX', 'ATR', 'CR', 'DMR', 'BR', 'TR']
    )
    m.species = Set(
        doc="chemical species",
        initialize=['CH4', 'H2O', 'O2', 'H2', 'CO', 'CO2']
    )
    m.syngas_tech_units = Set(
        doc="process units involved in syngas process technologies",
        initialize=['compressor', 'exchanger', 'reformer']
    )
    m.utilities = Set(
        doc="process utilities",
        initialize=['power', 'coolingwater', 'naturalgas']
    )
    m.aux_equipment = Set(
        doc="auxiliary process equipment to condition syngas",
        initialize=['absorber1', 'bypass1', 'WGS', 'flash', 'PSA', 'absorber2', 'bypass3', 'compressor', 'bypass4']
    )
    m.process_options = Set(
        doc="process superstructure options",
        initialize=m.syngas_techs | m.aux_equipment
    )
    m.extra_process_nodes = Set(
        doc="extra process nodes for mixers and splitters",
        initialize=['in', 'ms1', 'm1', 's1', 'ms2', 'ms3', 'm2', 'ms4', 's2']
    )
    m.all_unit_types = Set(
        doc="all process unit types",
        initialize=m.process_options | m.syngas_tech_units | m.extra_process_nodes
    )
    m.superstructure_nodes = Set(
        doc="nodes in the superstructure",
        initialize=m.process_options | m.extra_process_nodes
    )
    group1 = {'absorber1', 'bypass1', 'WGS'}
    group2 = {'compressor', 'bypass3'}
    group3 = {'absorber2', 'bypass4'}
    m.streams = Set(
        doc="superstructure potential streams",
        initialize=[(tech, 'ms1') for tech in m.syngas_techs]
        + [('in', tech) for tech in m.syngas_techs]
        + [('ms1', option) for option in group1]
        + [(option, 'm1') for option in group1]
        + [('m1', 'flash'), ('flash', 's1'), ('s1', 'PSA'), ('s1', 'ms2')]
        + [('ms2', option) for option in group2]
        + [(option, 'ms3') for option in group2]
        + [('ms3', option) for option in group3]
        + [(option, 'm2') for option in group3]
        + [('PSA', 'ms4'), ('ms4', 'ms3'), ('ms4', 's2'), ('s2', 'm1'), ('s2', 'ms1')]
    )

    """
    Parameters
    """

    m.flow_limit = Param(initialize=5)
    m.min_flow_division = Param(initialize=0.1, doc="minimum flow fraction into active unit")

    m.raw_material_cost = Param(
        m.species,
        doc="raw material cost [dollar·kmol-1]",
        initialize={'CH4': 2.6826, 'H2O': 0.18, 'O2': 0.7432, 'H2': 8, 'CO': 0, 'CO2': 1.8946}
    )

    feed_ratios = {
        ('SMR', 'H2O'): 3,
        ('POX', 'O2'): 0.5,
        ('ATR', 'H2O'): 1.43,
        ('ATR', 'O2'): 0.6,
        ('CR', 'H2O'): 2.5,
        ('CR', 'O2'): 0.19,
        ('DMR', 'CO2'): 1,
        ('BR', 'H2O'): 1.6,
        ('BR', 'CO2'): 0.8,
        ('TR', 'H2O'): 2.46,
        ('TR', 'CO2'): 1.3,
        ('TR', 'O2'): 0.47,
    }

    data_dict = pd.read_csv('syngas_conversion_data.txt', delimiter=r'\s+').fillna(0).stack().to_dict()
    m.syngas_conversion_factor = Param(m.species, m.syngas_techs, initialize=data_dict)
    m.co2_ratio = Param(doc="CO2-CO ratio of total carbon in final syngas", initialize=0.05)

    # Capital costs -->  cap = (p1*x + p2)*(B1 + B2*Fmf*Fp)
    m.p1 = Param(
        m.all_unit_types,
        doc="capital cost variable parameter",
        initialize={
            'compressor': 172.4, 'exchanger': 59.99, 'reformer': 67.64, 'absorber1': 314.1, 'bypass1': 0,
            'WGS': 314.1, 'flash': 314.1, 'PSA': 3.1863e+04, 'absorber2': 314.1, 'bypass3': 0},
        default=0
    )
    m.p2 = Param(
        m.all_unit_types,
        doc="capital cost fixed parameter",
        initialize={
            'compressor': 104300, 'exchanger': 187100, 'reformer': 480100, 'absorber1': 1.531e+04, 'bypass1': 0,
            'WGS': 1.531e+04, 'flash': 1.531e+04, 'PSA': 666.34e+03, 'absorber2': 1.531e+04, 'bypass3': 0},
        default=0
    )
    m.material_factor = Param(
        m.all_unit_types,
        doc="capital cost material factor",
        initialize={
            'compressor': 3.5, 'exchanger': 1.7, 'reformer': 4, 'absorber1': 1, 'bypass1': 0,
            'WGS': 1, 'flash': 1, 'PSA': 3.5, 'absorber2': 1, 'bypass3': 0},
        default=0
    )
    m.B1 = Param(
        m.all_unit_types,
        doc="bare module parameter 1",
        initialize={
            'compressor': 0, 'exchanger': 1.63, 'reformer': 0, 'absorber1': 2.25, 'bypass1': 0,
            'WGS': 1.49, 'flash': 2.25, 'PSA': 0, 'absorber2': 2.25, 'bypass3': 0},
        default=0
    )
    m.B2 = Param(
        m.all_unit_types,
        doc="bare module parameter 2",
        initialize={
            'compressor': 1, 'exchanger': 1.66, 'reformer': 1, 'absorber1': 1.82, 'bypass1': 0,
            'WGS': 1.52, 'flash': 1.82, 'PSA': 1, 'absorber2': 1.82, 'bypass3': 0},
        default=0
    )
    data_dict = pd.read_csv('syngas_pressure_factor_data.txt', delimiter=r'\s+').stack().to_dict()
    m.syngas_pressure_factor = Param(m.syngas_tech_units, m.syngas_techs, initialize=data_dict)

    m.utility_cost = Param(
        m.utilities,
        doc="dollars per kWh of utility u [dollar·KWh-1]",
        initialize={'power': 0.127, 'coolingwater': 0.01, 'naturalgas': 0.036}
    )
    m.utility_emission = Param(
        m.utilities,
        doc="CO2 emitted per kW [kgCO2·kWh-1]",
        initialize={'power': 0.44732, 'coolingwater': 0, 'naturalgas': 0.2674}
    )
    m.raw_material_emission = Param(
        m.species,
        doc="kg CO2 emitted per kmol of raw material [kgCO2·kmol-1]",
        initialize={'CH4': 11.7749, 'H2O': 0, 'O2': 10.9525, 'H2': 0, 'CO': 0, 'CO2': 0}
    )

    m.interest_rate = Param(initialize=0.1, doc="annualization index")
    m.project_years = Param(initialize=8, doc="annualization years")
    m.annualization_factor = Expression(
        expr=m.interest_rate * (1 + m.interest_rate)**m.project_years / ((1 + m.interest_rate)**m.project_years - 1))
    m.CEPCI2015 = Param(initialize=560, doc="equipment cost index 2015")
    m.CEPCI2001 = Param(initialize=397, doc="equipment cost index 2001")
    m.cost_index_ratio = Param(initialize=m.CEPCI2015 / m.CEPCI2001, doc="cost index ratio")

    data_dict = pd.read_csv('syngas_utility_data.txt', delimiter=r'\s+').stack().to_dict()
    m.syngas_tech_utility_rate = Param(
        m.utilities, m.syngas_techs, initialize=data_dict,
        doc="kWh of utility u per kmol of methane fed in syngas process i [kWh·kmol methane -1]")

    data_dict = pd.read_csv('syngas_num_units_data.txt', delimiter=r'\s+').stack().to_dict()
    m.syngas_tech_num_units = Param(
        m.syngas_tech_units, m.syngas_techs, initialize=data_dict,
        doc="number of units h in syngas process i")

    m.syngas_tech_exchanger_area = Param(
        m.syngas_techs,
        doc="total exchanger area in process i per kmol·h-1 methane fed [m2·h·kmol methane -1]",
        initialize={'SMR': 0.885917, 'POX': 0.153036, 'ATR': 0.260322, 'CR': 0.726294,
                    'DMR': 0.116814, 'BR': 0.825808, 'TR': 0.10539})

    m.syngas_tech_reformer_duty = Param(
        m.syngas_techs,
        doc="reformer duties per kmol methane fed [kWh·kmol methane-1]",
        initialize={'SMR': 54.654, 'POX': 39.0104, 'ATR': 44.4600, 'CR': 68.2382,
                    'DMR': 68.412, 'BR': 61.442, 'TR': 6.592})

    m.process_tech_pressure = Param(
        m.syngas_techs,
        doc="process i operating pressure  [bar]",
        initialize={'SMR': 20, 'POX': 30, 'ATR': 25, 'CR': 25, 'DMR': 1, 'BR': 7, 'TR': 20}
    )
    m.final_syngas_pressure = Param(doc="final syngas pressure [bar]", initialize=1)
    m.psa_hydrogen_recovery = Param(initialize=0.9, doc="percentage of hydrogen separated from the inlet syngas stream")
    m.psa_separation_hydrogen_purity = Param(initialize=0.999)

    m.Keqw = Param(initialize=83.3429, doc="equilibrium constant for WGS reaction at 250ºC")

    m.stoichiometric_number = Param(doc="stoichiometric number of product syngas", initialize=1)
    m.max_impurity = Param(initialize=0.1, doc="maximum allowed of impurities")
    m.max_syngas_techs = Param(initialize=1, doc="Number of syngas technologies that can be selected")

    """
    Variables
    """

    m.flow = Var(m.streams, m.species, bounds=(0, m.flow_limit))
    m.wgs_steam = Var(doc="steam molar flow provided in the WGS reactor [kmol·s-1]", bounds=(0, None))
    m.oxygen_flow = Var(doc="O2 molar flow provided in the selective oxidation reactor [kmol·s-1]", bounds=(0, None))
    m.Fabs1 = Var(bounds=(0, None), doc="molar flow of CO2 absorbed in absorber1 [kmol·s-1]")
    m.Fabs2 = Var(bounds=(0, None), doc="molar flow of CO2 absorbed in absorber2 [kmol·s-1]")
    m.flash_water = Var(bounds=(0, None), doc="water removed in flash [kmol·s-1]")
    m.co2_inject = Var(bounds=(0, None), doc="molar flow of CO2 used to adjust syngas composition [kmol·s-1]")
    m.psa_recovered = Var(m.species, bounds=(0, None), doc="pure hydrogen stream retained in the PSA [kmol·s-1]")
    m.purge_flow = Var(m.species, bounds=(0, None), doc="purged molar flow from the PSA [kmol·s-1]")
    m.final_syngas_flow = Var(m.species, bounds=(0, m.flow_limit), doc="final adjusted syngas molar flow [kmol·s-1]")

    @m.Expression(m.superstructure_nodes, m.species)
    def flow_into(m, option, species):
        return sum(m.flow[src, sink, species] for src, sink in m.streams if sink == option)

    @m.Expression(m.superstructure_nodes, m.species)
    def flow_out_from(m, option, species):
        return sum(m.flow[src, sink, species] for src, sink in m.streams if src == option)

    @m.Expression(m.superstructure_nodes)
    def total_flow_into(m, option):
        return sum(m.flow_into[option, species] for species in m.species)

    @m.Expression(m.superstructure_nodes)
    def total_flow_from(m, option):
        return sum(m.flow_out_from[option, species] for species in m.species)

    m.base_tech_capital_cost = Var(m.syngas_techs, m.syngas_tech_units, bounds=(0, None))
    m.base_tech_operating_cost = Var(m.syngas_techs, m.utilities, bounds=(0, None))
    m.raw_material_total_cost = Var(bounds=(0, None), doc="total cost of raw materials [$·s-1]")

    @m.Expression(m.species)
    def raw_material_flow(m, species):
        return sum(m.flow['in', tech, species] for tech in m.syngas_techs)

    m.syngas_tech_cost = Var(bounds=(0, None), doc="total cost of sygas process [$·y-1]")
    m.syngas_tech_emissions = Var(bounds=(0, None), doc="CO2 emission of syngas processes")

    @m.Expression(m.syngas_techs, m.syngas_tech_units)
    def module_factors(m, tech, equip):
        return m.B1[equip] + m.B2[equip] * m.material_factor[equip] * m.syngas_pressure_factor[equip, tech]

    @m.Expression(m.syngas_techs, m.syngas_tech_units)
    def variable_utilization(m, tech, equip):
        variable_rate_term = {'compressor': m.syngas_tech_utility_rate['power', tech],
                              'exchanger': m.syngas_tech_exchanger_area[tech],
                              'reformer': m.syngas_tech_reformer_duty[tech]}
        return variable_rate_term[equip] * m.flow['in', tech, 'CH4'] * 3600

    m.aux_unit_capital_cost = Var(m.aux_equipment, bounds=(0, None), doc="auxiliary unit capital cost [$·h-1]")

    m.first_stage_outlet_pressure = Var(doc="final pressure in the mixer before WGS and absorber", bounds=(0, None))
    m.syngas_tech_outlet_pressure = Var(m.syngas_techs, doc="final pressure after compression after syngas synthesis [bar]", bounds=(0, None))
    m.syngas_tech_compressor_power = Var(m.syngas_techs, doc="utility of compressor i after syngas synthesis", bounds=(0, None))
    m.syngas_tech_compressor_cost = Var(doc="capital cost of compressors after syngas synthesis", bounds=(0, None))

    m.Xw = Var(bounds=(0, None), doc="Moles per second reacted in WGS reactor")
    m.wgs_inlet_temperature = Var(bounds=(0, None), doc="WGS reactor temperature before adjustment [ºC]")
    m.wgs_heater = Var(bounds=(0, None), doc="duty required to preheat the syngas molar flow entering the WGS reactor [kW]")
    m.psa_power = Var(bounds=(0, None), doc="power consumed by the PSA [kWh]")
    m.syngas_power = Var(bounds=(0, None), doc="power needed to compress syngas from Pinlet to 30.01 bar")

    m.syngas_total_flow = Expression(expr=sum(m.final_syngas_flow[species] for species in {'H2', 'CO2', 'CO', 'CH4'}))

    @m.Expression(m.aux_equipment)
    def aux_module_factors(m, equip):
        return m.B1[equip] + m.B2[equip] * m.material_factor[equip]

    m.final_total_emissions = Expression(
        expr=m.syngas_tech_emissions*3600
        + m.Fabs1*3600*44 + m.Fabs2*3600*44
        - m.co2_inject*3600*44 + m.oxygen_flow*m.raw_material_emission['O2']*3600
        + m.wgs_steam*m.raw_material_emission['H2O']*3600
        + m.purge_flow['CO2']*3600*44
        + (m.psa_power + m.syngas_power + sum(m.syngas_tech_compressor_power[tech] for tech in m.syngas_techs)
           )*m.utility_emission['power']
        + m.wgs_heater*0.094316389565193)

    m.final_total_cost = Expression(
        expr=m.syngas_tech_cost*3600 + m.syngas_tech_compressor_cost
        + sum(m.aux_unit_capital_cost[option] for option in m.aux_equipment) * m.annualization_factor
        + (m.psa_power + m.syngas_power + sum(m.syngas_tech_compressor_power[tech] for tech in m.syngas_techs)
           )*m.utility_cost['power']
        + m.wgs_heater*0.064)

    """
    Constraints
    """

    @m.Constraint(m.syngas_techs, m.species)
    def syngas_process_feed_species_ratio(m, tech, species):
        if species == 'CH4':
            return Constraint.Skip
        return m.flow['in', tech, species] == feed_ratios.get((tech, species), 0) * m.flow['in', tech, 'CH4']

    @m.Constraint(m.syngas_techs, m.species)
    def syngas_conversion_calc(m, tech, species):
        return m.flow[tech, 'ms1', species] == m.flow['in', tech, 'CH4'] * m.syngas_conversion_factor[species, tech]

    m.raw_material_cost_calc = Constraint(expr=m.raw_material_total_cost == (
        sum(m.raw_material_flow[species] * m.raw_material_cost[species] for species in m.species)
        + m.wgs_steam * m.raw_material_cost['H2O'] + m.oxygen_flow * m.raw_material_cost['O2']
    ))

    @m.Disjunct(m.process_options)
    def unit_exists(disj, option):
        pass

    @m.Disjunct(m.process_options)
    def unit_absent(no_unit, option):
        @no_unit.Constraint(m.species)
        def no_flow_in(disj, species):
            return m.flow_into[option, species] == 0

        @no_unit.Constraint(m.species)
        def no_flow_out(disj, species):
            return m.flow_out_from[option, species] == 0

    @m.Disjunction(m.process_options)
    def unit_exists_or_not(disj, option):
        return [m.unit_exists[option], m.unit_absent[option]]

    m.Yunit = BooleanVar(m.process_options, doc="Boolean variable for existence of a process unit")
    for option in m.process_options:
        m.Yunit[option].set_binary_var(m.unit_exists[option].indicator_var)

    for tech in m.syngas_techs:
        # Capital costs -->  cap = (p1*x + p2)*(B1 + B2*Fmf*Fp)
        tech_selected = m.unit_exists[tech]

        @tech_selected.Constraint(m.syngas_tech_units)
        def base_tech_capital_cost_calc(disj, equip):
            return m.base_tech_capital_cost[tech, equip] == (
                (m.p1[equip] * m.variable_utilization[tech, equip]
                 + m.p2[equip] * m.syngas_tech_num_units[equip, tech] * m.module_factors[tech, equip])
                * m.cost_index_ratio / 8000
            )

        @tech_selected.Constraint(m.utilities)
        def base_tech_operating_cost_calc(disj, util):
            return m.base_tech_operating_cost[tech, util] == (
                m.syngas_tech_utility_rate[util, tech] * m.flow['in', tech, 'CH4'] * 3600 * m.utility_cost[util]
            )

    m.syngas_process_cost_calc = Constraint(expr=m.syngas_tech_cost == (
        sum(sum(m.base_tech_capital_cost[tech, equip] for equip in m.syngas_tech_units) * m.annualization_factor
            + sum(m.base_tech_operating_cost[tech, util] for util in m.utilities)
            for tech in m.syngas_techs) + m.raw_material_total_cost * 3600) / 3600)

    m.syngas_emissions_calc = Constraint(expr=m.syngas_tech_emissions == (
        # Emissions from utilities
        sum(m.base_tech_operating_cost[tech, util] / m.utility_cost[util] * m.utility_emission[util]
            for tech in m.syngas_techs for util in m.utilities)
        # CO2 consumed by syngas processes
        - sum(m.flow['in', tech, 'CO2'] for tech in m.syngas_techs) * 3600 * 44
        # Emissions from raw materials
        + sum(m.raw_material_flow[species] * m.raw_material_emission[species] * 3600 for species in m.species)) / 3600)

    # Syngas process pressure adjustment

    @m.Disjunct(m.syngas_techs)
    def stage_one_compressor(disj, tech):
        disj.compressor_power_calc = Constraint(expr=m.syngas_tech_compressor_power[tech] == (
            (1.5 / (1.5 - 1)) / 0.8 * (40 + 273) * 8.314 * sum(m.flow[tech, 'ms1', species] for species in m.species)
            * ((m.syngas_tech_outlet_pressure[tech] / m.process_tech_pressure[tech]) ** (1.5 - 1 / 1.5) - 1)
        ))
        pass

    @m.Disjunct(m.syngas_techs)
    def stage_one_bypass(bypass, tech):
        bypass.no_pressure_increase = Constraint(expr=m.syngas_tech_outlet_pressure[tech] == m.process_tech_pressure[tech])
        pass

    @m.Disjunction(m.syngas_techs)
    def stage_one_compressor_or_bypass(m, tech):
        return [m.stage_one_compressor[tech], m.stage_one_bypass[tech]]

    m.Ycomp = BooleanVar(m.syngas_techs)
    for tech in m.syngas_techs:
        m.Ycomp[tech].set_binary_var(m.stage_one_compressor[tech].indicator_var)

    @m.LogicalStatement(m.syngas_techs)
    def compressor_implies_tech(m, tech):
        return m.Ycomp[tech].implies(m.Yunit[tech])

    m.syngas_tech_compressor_cost_calc = Constraint(expr=m.syngas_tech_compressor_cost == (
        ((3.553 * 10 ** 5) * sum(m.Ycomp[tech].as_binary() for tech in m.syngas_techs)
         + 586 * sum(m.syngas_tech_compressor_power[tech] for tech in m.syngas_techs))
        * m.annualization_factor / 8000
    ))

    for tech in m.syngas_techs:
        tech_selected = m.unit_exists[tech]
        tech_selected.pressure_balance = Constraint(expr=m.first_stage_outlet_pressure <= m.syngas_tech_outlet_pressure[tech])

    # ms1 balances
    @m.Constraint(m.species)
    def ms1_mass_balance(m, species):
        return m.flow_into['ms1', species] == m.flow_out_from['ms1', species]

    for this_unit in {'absorber1', 'WGS'}:
        unit_exists = m.unit_exists[this_unit]

        @unit_exists.Constraint(m.species)
        def unit_inlet_composition_balance(disj, species):
            total_flow = sum(m.flow_out_from['ms1', jj] for jj in m.species)
            total_flow_to_this_unit = sum(m.flow_into[this_unit, jj] for jj in m.species)
            total_flow_species = m.flow_out_from['ms1', species]
            return total_flow * m.flow['ms1', this_unit, species] == total_flow_to_this_unit * total_flow_species

    # WGS Reactor  CO + H2O <-> CO2 + H2
    wgs_exists = m.unit_exists['WGS']
    wgs_exists.CH4_balance = Constraint(expr=m.flow_out_from['WGS', 'CH4'] == m.flow_into['WGS', 'CH4'])
    wgs_exists.CO_balance = Constraint(expr=m.flow_out_from['WGS', 'CO'] == m.flow_into['WGS', 'CO'] - m.Xw)
    wgs_exists.CO2_balance = Constraint(expr=m.flow_out_from['WGS', 'CO2'] == m.flow_into['WGS', 'CO2'] + m.Xw)
    wgs_exists.H2_balance = Constraint(expr=m.flow_out_from['WGS', 'H2'] == m.flow_into['WGS', 'H2'] + m.Xw)
    wgs_exists.H2O_balance = Constraint(expr=m.flow_out_from['WGS', 'H2O'] == m.flow_into['WGS', 'H2O'] + m.wgs_steam - m.Xw)
    wgs_exists.max_molar_reaction = Constraint(expr=m.Xw <= m.flow_into['WGS', 'CO'])

    wgs_exists.rxn_equilibrium = Constraint(
        expr=m.Keqw * m.flow_out_from['WGS', 'CO'] * m.flow_out_from['WGS', 'H2O'] == (
            m.flow_out_from['WGS', 'CO2'] * m.flow_out_from['WGS', 'H2']
        ))

    wgs_exists.capital_cost = Constraint(expr=m.aux_unit_capital_cost['WGS'] == (
        (m.p1['WGS'] * 100 * m.flow_out_from['WGS', 'H2'] + m.p2['WGS'])
        * m.aux_module_factors['WGS'] / 8000 * m.cost_index_ratio
    ))

    wgs_exists.temperature_balance = Constraint(
        expr=m.wgs_inlet_temperature * m.total_flow_into['WGS'] == (
            sum(m.flow[tech, 'ms1', species] for tech in m.syngas_techs for species in m.species) * 250
            + sum(m.flow['s2', 'ms1', species] for species in m.species) * 40
        ))
    wgs_exists.heater_duty = Constraint(
        expr=m.wgs_heater == m.total_flow_into['WGS'] * 46 * (250 - m.wgs_inlet_temperature))

    # Bypass 1
    bypass1_exists = m.unit_exists['bypass1']

    @bypass1_exists.Constraint(m.species)
    def bypass1_mass_balance(disj, species):
        return m.flow_into['bypass1', species] == m.flow_out_from['bypass1', species]

    # Absorber
    absorber_exists = m.unit_exists['absorber1']

    @absorber_exists.Constraint(m.species)
    def absorber_mass_balance(disj, species):
        return m.flow_out_from['absorber1', species] == m.flow_into['absorber1', species] - (
            m.Fabs1 if species == 'CO2' else 0)

    absorber_exists.co2_absorption = Constraint(expr=m.Fabs1 == 0.96 * m.flow_into['absorber1', 'CO2'])
    absorber_exists.cost = Constraint(expr=m.aux_unit_capital_cost['absorber1'] == (
        (m.p1['absorber1'] * 100 * m.Fabs1 + m.p2['absorber1'])
        * m.aux_module_factors['absorber1'] / 8000 * m.cost_index_ratio
    ))

    m.unit_absent['absorber1'].no_absorption = Constraint(expr=m.Fabs1 == 0)

    for this_unit in group1:
        unit_exists = m.unit_exists[this_unit]
        unit_exists.minimum_flow = Constraint(expr=m.total_flow_into[this_unit] >= m.total_flow_from['ms1'] * m.min_flow_division)

    # Flash
    @m.Constraint(m.species)
    def m1_mass_balance(m, species):
        return m.flow_into['m1', species] == m.flow_out_from['m1', species]

    flash_exists = m.unit_exists['flash']

    @flash_exists.Constraint(m.species)
    def flash_mass_balance(disj, species):
        return m.flow_out_from['flash', species] == (m.flow_into['flash', species] if not species == 'H2O' else 0)

    flash_exists.water_sep = Constraint(expr=m.flash_water == m.flow_into['flash', 'H2O'])

    @m.Constraint(m.species)
    def post_flash_split_outlet(m, species):
        return m.flow_out_from['flash', species] == m.flow_into['PSA', species] + m.flow_into['ms2', species]

    flash_exists.cost = Constraint(expr=m.aux_unit_capital_cost['flash'] == (
        (m.p1['flash'] * 100 * m.flash_water + m.p2['flash'])
        * m.aux_module_factors['flash'] / 8000 * m.cost_index_ratio
    ))

    # PSA
    psa_exists = m.unit_exists['PSA']

    @psa_exists.Constraint(m.species)
    def psa_inlet_composition_balance(disj, species):
        total_flow = sum(m.flow_out_from['s1', jj] for jj in m.species)
        total_flow_to_this_unit = sum(m.flow_into['PSA', jj] for jj in m.species)
        total_flow_species = m.flow_out_from['s1', species]
        return total_flow * m.flow['s1', 'PSA', species] == total_flow_to_this_unit * total_flow_species

    @m.Constraint(m.species)
    def ms2_inlet_composition_balance(disj, species):
        total_flow = sum(m.flow_out_from['s1', jj] for jj in m.species)
        total_flow_to_this_unit = sum(m.flow_into['ms2', jj] for jj in m.species)
        total_flow_species = m.flow_out_from['s1', species]
        return total_flow * m.flow['s1', 'ms2', species] == total_flow_to_this_unit * total_flow_species

    @psa_exists.Constraint(m.species)
    def psa_mass_balance(disj, species):
        return m.flow_out_from['PSA', species] + m.psa_recovered[species] == m.flow_into['PSA', species]

    @psa_exists.Constraint(m.species)
    def psa_recovery(disj, species):
        return m.flow_out_from['PSA', species] == m.flow_into['PSA', species] * (
            (1 - m.psa_hydrogen_recovery if species == 'H2' else m.psa_separation_hydrogen_purity)
        )

    psa_exists.cost = Constraint(expr=m.aux_unit_capital_cost['PSA'] == (
        (m.p1['PSA'] * m.flow_into['PSA', 'H2'] + m.p2['PSA'])
        * m.aux_module_factors['PSA'] / 8000
    ))
    psa_exists.psa_utility = Constraint(expr=m.psa_power*m.first_stage_outlet_pressure**(1.5-1/1.5) == (
        (1.5/(1.5-1))/0.8*(40+273) * 8.314 * m.total_flow_into['PSA']
        * ((30+1e-6)**(1.5-1/1.5) - m.first_stage_outlet_pressure**(1.5-1/1.5))))

    psa_absent = m.unit_absent['PSA']

    @psa_absent.Constraint(m.species)
    def no_psa_recovery(disj, species):
        return m.psa_recovered[species] == 0

    @psa_absent.Constraint(m.species)
    def no_purge(disj, species):
        return m.purge_flow[species] == 0

    # ms4
    @m.Constraint(m.species)
    def ms4_inlet_mass_balance(m, species):
        return m.flow_out_from['PSA', species] == m.flow_into['ms4', species]

    @m.Constraint(m.species)
    def ms4_outlet_mass_balance(m, species):
        return m.flow_out_from['ms4', species] + m.purge_flow[species] == m.flow_into['ms4', species]

    @m.Constraint(m.species)
    def purge_flow_limit(m, species):
        return m.purge_flow[species] <= m.flow_into['ms4', species] * 0.01

    @m.Constraint(m.species)
    def s2_inlet_composition(m, species):
        total_flow = sum(m.flow_into['ms4', jj] for jj in m.species)
        total_flow_to_this_unit = sum(m.flow_into['s2', jj] for jj in m.species)
        total_flow_species = m.flow_into['ms4', species]
        return total_flow * m.flow['ms4', 's2', species] == total_flow_to_this_unit * total_flow_species

    @m.Constraint(m.species)
    def ms4_to_ms3_composition(m, species):
        total_flow = sum(m.flow_into['ms4', jj] for jj in m.species)
        total_flow_to_this_unit = sum(m.flow['ms4', 's2', jj] for jj in m.species)
        total_flow_species = m.flow_into['ms4', species]
        return total_flow * m.flow['ms4', 's2', species] == total_flow_to_this_unit * total_flow_species

    # s2
    @m.Constraint(m.species)
    def s2_mass_balance(m, species):
        return m.flow_into['s2', species] == m.flow['s2', 'ms1', species]

    @m.Constraint(m.species)
    def no_flow_s2_to_m1(m, species):
        return m.flow['s2', 'm1', species] == 0

    # ms2
    @m.Constraint(m.species)
    def ms2_mass_balance(m, species):
        return m.flow_out_from['ms2', species] == (
                m.flow_into['ms2', species] + (m.co2_inject if species == 'CO2' else 0))

    # bypass3
    bypass3_exists = m.unit_exists['bypass3']

    @bypass3_exists.Constraint(m.species)
    def bypass3_mass_balance(disj, species):
        return m.flow_into['bypass3', species] == m.flow_out_from['bypass3', species]

    # compressor
    compressor_exists = m.unit_exists['compressor']

    @compressor_exists.Constraint(m.species)
    def compressor_inlet_mass_balance(disj, species):
        return m.flow_into['compressor', species] == m.flow_out_from['ms2', species]

    @compressor_exists.Constraint(m.species)
    def compressor_mass_balance(disj, species):
        return m.flow_out_from['compressor', species] == m.flow_into['compressor', species]

    compressor_exists.cost = Constraint(expr=m.aux_unit_capital_cost['compressor'] == (
        ((3.553 * 10 ** 5) + 586 * m.syngas_power) / 8000 * m.cost_index_ratio
    ))

    compressor_exists.work = Constraint(expr=m.syngas_power * m.first_stage_outlet_pressure**(1.5-1/1.5) == (
        (1.5/(1.5-1))/0.8*(40+273)*8.314*m.total_flow_into['compressor']
        * (m.final_syngas_pressure**(1.5-1/1.5) - m.first_stage_outlet_pressure**(1.5-1/1.5))))

    no_compressor = m.unit_absent['compressor']
    no_compressor.final_pressure = Constraint(expr=m.first_stage_outlet_pressure >= m.final_syngas_pressure)

    compressor_exists.compressor_minimum_flow = Constraint(expr=m.total_flow_into['compressor'] >= (
        m.total_flow_from['flash'] * m.min_flow_division
    ))
    psa_exists.psa_minimum_flow = Constraint(expr=m.total_flow_into['PSA'] >= (
        m.total_flow_from['flash'] * m.min_flow_division
    ))

    m.compressor_or_bypass = LogicalStatement(expr=Exactly(1, m.Yunit['bypass3'], m.Yunit['compressor']))

    # ms3
    @m.Constraint(m.species)
    def ms3_mass_balance(m, species):
        return m.flow_into['ms3', species] == m.flow_out_from['ms3', species]

    # bypass 4
    bypass4_exists = m.unit_exists['bypass4']

    @bypass4_exists.Constraint(m.species)
    def bypass4_mass_balance(disj, species):
        return m.flow_into['bypass4', species] == m.flow_out_from['bypass4', species]

    # absorber 2
    absorber2_exists = m.unit_exists['absorber2']
    
    @absorber2_exists.Constraint(m.species)
    def absorber2_mass_balance(disj, species):
        return m.flow_out_from['absorber2', species] == (
            m.flow_into['absorber2', species] - (m.Fabs2 if species == 'CO2' else 0))
    
    absorber2_exists.co2_absorption = Constraint(expr=m.Fabs2 == 0.96 * m.flow_into['absorber2', 'CO2'])
    absorber2_exists.cost = Constraint(expr=m.aux_unit_capital_cost['absorber2'] == (
        (m.p1['absorber2'] * 100 * m.Fabs2 + m.p2['absorber2'])
        * m.aux_module_factors['absorber2'] / 8000 * m.cost_index_ratio
    ))

    m.unit_absent['absorber2'].no_absorption = Constraint(expr=m.Fabs2 == 0)
    
    m.only_one_absorber = LogicalStatement(expr=AtMost(1, m.Yunit['absorber1'], m.Yunit['absorber2']))

    @m.Constraint(m.species)
    def final_mass_balance(m, species):
        return m.final_syngas_flow[species] == m.flow_into['m2', species]

    m.syngas_stoich_number = Constraint(
        expr=m.stoichiometric_number * (m.final_syngas_flow['CO'] + m.final_syngas_flow['CO2']) == (
            m.final_syngas_flow['H2'] - m.final_syngas_flow['CO2']
        ))
    m.impurity_limit = Constraint(
        expr=m.max_impurity * sum(m.final_syngas_flow[species] for species in {'CO', 'H2', 'CH4', 'CO2', 'H2O'})
        >= sum(m.final_syngas_flow[species] for species in {'CH4', 'H2O'})
    )

    m.syngas_process_limits = LogicalStatement(
        expr=AtMost(m.max_syngas_techs, [m.Yunit[tech] for tech in m.syngas_techs]))

    # Bounds
    m.wgs_heater.setub(10000)
    m.syngas_tech_compressor_power.setub(10000)
    m.syngas_tech_compressor_cost.setub(10000)
    m.psa_power.setub(10000)

    m.Fabs1.setub(20)
    m.Fabs2.setub(20)
    m.flash_water.setub(20)

    m.psa_recovered.setub(20)
    m.purge_flow.setub(20)
    m.co2_inject.setub(20)
    m.final_syngas_flow.setub(20)
    m.syngas_power.setub(10000)

    m.raw_material_total_cost.setub(1000)
    m.base_tech_capital_cost.setub(10000)
    m.base_tech_operating_cost.setub(10000)
    m.syngas_tech_cost.setub(1000)
    m.syngas_tech_emissions.setub(100)
    m.syngas_tech_emissions.setlb(-100)

    m.Xw.setub(30)
    m.wgs_inlet_temperature.setub(250)

    m.syngas_tech_outlet_pressure.setub(100)
    m.first_stage_outlet_pressure.setlb(1)
    m.first_stage_outlet_pressure.setub(m.final_syngas_pressure)

    m.wgs_steam.setub(20)
    m.oxygen_flow.setub(0)
    # Not done yet:
    # FPSA.up(j) = 20;
    # FPSAmain.up(j) = 20;
    m.flow['ms4', 's2', :].setub(20)
    m.flow['s2', 'ms1', :].setub(20)
    m.flow['ms4', 'ms3', :].setub(20)
    m.flow['psa', 'ms4', :].setub(20)

    m.no_downstream_oxygen = Constraint(expr=m.flow_out_from['ms1', 'O2'] == 0)

    # No oxygen in downstream processes: not enforced yet

    m.aux_unit_capital_cost.setub(10000)

    m.always_use_flash = LogicalStatement(expr=m.Yunit['flash'])

    m.syngas_minimum_demand = Constraint(expr=m.syngas_total_flow >= 0.3)
    m.syngas_maximum_demand = Constraint(expr=m.syngas_total_flow <= 5)

    m.final_syngas_flow['CO'].fix(0.3)
    m.final_syngas_flow['CO2'].fix(0.3*m.co2_ratio)

    m.obj = Objective(expr=m.final_total_cost)

    # Attempting to troubleshoot:
    # m.optimal_feed = Constraint(expr=m.flow_out_from['in', 'CH4'] == 0.1674)
    # m.wgs_flow = Constraint(expr=(0.00991008 - 1E-5, m.flow['ms1', 'WGS', 'CH4'], 0.00991008 + 1e-5))
    # m.wgs_conv = Constraint(expr=(0.0201 - 1E-4, m.Xw, 0.0201 + 1E-4))
    # m.use_dmr = LogicalStatement(
    #     expr=m.Yunit['DMR']
    #          & m.Yunit['WGS'] & ~m.Yunit['bypass1'] & ~m.Yunit['absorber1']
    #          & ~m.Yunit['PSA']
    #          & m.Yunit['absorber2']
    #          & m.Yunit['bypass4']
    #          & m.Yunit['bypass3']
    # )
    # psa_exists.psa_minimum_flow.deactivate()
    # m.flow['s1', 'PSA', 'CO'].fix(0)
    # m.post_flash_split_outlet.deactivate()

    return m


def display_nonzeros(var):
    if var.is_indexed():
        def nonzero_rows():
            for k, v in var.items():
                if v.value == 0:
                    continue
                yield k, [value(v.lb), v.value, value(v.ub), v.fixed, v.stale, v.domain]
        _attr, _, _header, _ = var._pprint()
        var._pprint_base_impl(
            None, False, "", var.local_name, var.doc,
            var.is_constructed(), _attr, nonzero_rows(), _header, lambda k, v: v)
    else:
        var.display()


if __name__ == "__main__":
    m = build_model()
    TransformationFactory('core.logical_to_linear').apply_to(m)
    # TransformationFactory('gdp.bigm').apply_to(m)
    # result = SolverFactory('gams').solve(
    #     m, tee=True, keepfiles=True, symbolic_solver_labels=True,
    #     solver='scip', add_options=['option optcr=0;'],
    #     # solver="cplex", add_options=['GAMS_MODEL.optfile=1;', '$onecho > cplex.opt', 'iis=1', '$offecho'],
    # )
    result = SolverFactory('gdpopt').solve(
        m, tee=True, mip_solver='gams',
        nlp_solver='gams', nlp_solver_args=dict(solver='scip', add_options=['option optcr=0;']),
        minlp_solver='gams', minlp_solver_args=dict(solver='baron', add_options=['option optcr=0;'])
    )
    if not result.solver.termination_condition == tc.optimal:
        print("Termination Condition: ", result.solver.termination_condition)
    else:
        update_boolean_vars_from_binary(m)
        display_nonzeros(m.flow)
        # m.flow.display()
        m.Yunit.display()
        m.final_total_cost.display()
        m.final_total_emissions.display()



import re


# Alphanumeric sort, taken from https://nedbatchelder.com/blog/200712/human_sorting.html
def tryint(s):
    try:
        return int(s)
    except ValueError:
        return s


def alphanum_key(s):
    """ Turn a string into a list of string and number chunks.
        "z23a" -> ["z", 23, "a"]
    """
    return [tryint(c) for c in re.split('([0-9]+)', s)]


def alphanum_sorted(l):
    """ Sort the given list in the way that humans expect.
    """
    return sorted(l, key=alphanum_key)



"""Heat integration case study.

This is example 1 of the Yee & Grossmann, 1990 paper "Simultaneous optimization
models for heat integration--II".
DOI: 10.1016/0098-1354(90)85010-8

This file provides common modeling elements.

"""
from __future__ import division

from pyomo.environ import (
    ConcreteModel, Constraint, minimize, NonNegativeReals, Objective, Param, RangeSet, Set, Suffix, value, Var, )
from pyomo.gdp import Disjunct, Disjunction

from .cafaro_approx import calculate_cafaro_coefficients


def build_model(use_cafaro_approximation, num_stages):
    """Build the model."""
    m = ConcreteModel()
    m.hot_process_streams = Set(initialize=['H1', 'H2'])
    m.cold_process_streams = Set(initialize=['C1', 'C2'])
    m.process_streams = m.hot_process_streams | m.cold_process_streams
    m.hot_utility_streams = Set(initialize=['steam'])
    m.cold_utility_streams = Set(initialize=['water'])
    m.hot_streams = Set(
        initialize=m.hot_process_streams | m.hot_utility_streams)
    m.cold_streams = Set(
        initialize=m.cold_process_streams | m.cold_utility_streams)
    m.utility_streams = Set(
        initialize=m.hot_utility_streams | m.cold_utility_streams)
    m.streams = Set(
        initialize=m.process_streams | m.utility_streams)
    m.valid_matches = Set(
        initialize=(m.hot_process_streams * m.cold_streams) |
        (m.hot_utility_streams * m.cold_process_streams),
        doc="Match all hot streams to cold streams, but exclude "
        "matches between hot and cold utilities.")
    # m.EMAT = Param(doc="Exchanger minimum approach temperature [K]",
    #                initialize=1)
    # Unused right now, but could be used for variable bound tightening
    # in the LMTD calculation.

    m.stages = RangeSet(num_stages)

    m.T_in = Param(
        m.streams, doc="Inlet temperature of stream [K]",
        initialize={'H1': 443,
                    'H2': 423,
                    'C1': 293,
                    'C2': 353,
                    'steam': 450,
                    'water': 293})
    m.T_out = Param(
        m.streams, doc="Outlet temperature of stream [K]",
        initialize={'H1': 333,
                    'H2': 303,
                    'C1': 408,
                    'C2': 413,
                    'steam': 450,
                    'water': 313})

    m.heat_exchanged = Var(
        m.valid_matches, m.stages,
        domain=NonNegativeReals,
        doc="Heat exchanged from hot stream to cold stream in stage [kW]",
        initialize=1, bounds=(0, 5000))

    m.overall_FCp = Param(
        m.process_streams,
        doc="Flow times heat capacity of stream [kW / K]",
        initialize={'H1': 30,
                    'H2': 15,
                    'C1': 20,
                    'C2': 40})
    m.utility_usage = Var(
        m.utility_streams,
        doc="Hot or cold utility used [kW]",
        domain=NonNegativeReals, initialize=1, bounds=(0, 5000))
    m.stage_entry_T = Var(
        m.streams, m.stages,
        doc="Temperature of stream at stage entry.",
        initialize=350,
        bounds=(293, 450)  # TODO set to be equal to min and max temps
    )
    m.stage_exit_T = Var(
        m.streams, m.stages,
        doc="Temperature of stream at stage exit.",
        initialize=350,
        bounds=(293, 450)  # TODO set to be equal to min and max temps
    )
    # Improve bounds on stage entry and exit temperatures
    for strm, stg in m.process_streams * m.stages:
        m.stage_entry_T[strm, stg].setlb(
            min(value(m.T_in[strm]), value(m.T_out[strm])))
        m.stage_exit_T[strm, stg].setlb(
            min(value(m.T_in[strm]), value(m.T_out[strm])))
        m.stage_entry_T[strm, stg].setub(
            max(value(m.T_in[strm]), value(m.T_out[strm])))
        m.stage_exit_T[strm, stg].setub(
            max(value(m.T_in[strm]), value(m.T_out[strm])))
    for strm, stg in m.utility_streams * m.stages:
        _fix_and_bound(m.stage_entry_T[strm, stg], m.T_in[strm])
        _fix_and_bound(m.stage_exit_T[strm, stg], m.T_out[strm])
    for strm in m.hot_process_streams:
        _fix_and_bound(m.stage_entry_T[strm, 1], m.T_in[strm])
        _fix_and_bound(m.stage_exit_T[strm, num_stages], m.T_out[strm])
    for strm in m.cold_process_streams:
        _fix_and_bound(m.stage_exit_T[strm, 1], m.T_out[strm])
        _fix_and_bound(m.stage_entry_T[strm, num_stages], m.T_in[strm])

    m.BigM = Suffix(direction=Suffix.LOCAL)

    m.utility_unit_cost = Param(
        m.utility_streams,
        doc="Annual unit cost of utilities [$/kW]",
        initialize={'steam': 80, 'water': 20})

    m.module_sizes = Set(initialize=[10, 50, 100])
    m.max_num_modules = Param(m.module_sizes, initialize={
        # 5: 100,
        10: 50,
        50: 10,
        100: 5,
        # 250: 2
    }, doc="maximum number of each module size available.")

    m.exchanger_fixed_unit_cost = Param(
        m.valid_matches, default=2000)
    m.exchanger_area_cost_factor = Param(
        m.valid_matches, default=1000,
        initialize={
            ('steam', cold): 1200
            for cold in m.cold_process_streams},
        doc="1200 for heaters. 1000 for all other exchangers.")
    m.area_cost_exponent = Param(default=0.6)

    if use_cafaro_approximation:
        k, b = calculate_cafaro_coefficients(10, 500, m.area_cost_exponent)
        m.cafaro_k = Param(default=k)
        m.cafaro_b = Param(default=b)

    @m.Param(m.valid_matches, m.module_sizes,
             doc="Area cost factor for modular exchangers.")
    def module_area_cost_factor(m, hot, cold, area):
        if hot == 'steam':
            return 1300
        else:
            return 1100

    m.module_fixed_unit_cost = Param(default=0)
    m.module_area_cost_exponent = Param(default=0.6)

    @m.Param(m.valid_matches, m.module_sizes,
             doc="Cost of a module with a particular area.")
    def module_area_cost(m, hot, cold, area):
        return (m.module_area_cost_factor[hot, cold, area]
                * area ** m.module_area_cost_exponent)

    m.U = Param(
        m.valid_matches,
        default=0.8,
        initialize={
            ('steam', cold): 1.2
            for cold in m.cold_process_streams},
        doc="Overall heat transfer coefficient."
        "1.2 for heaters. 0.8 for everything else.")

    m.exchanger_hot_side_approach_T = Var(
        m.valid_matches, m.stages,
        doc="Temperature difference between the hot stream inlet and cold "
        "stream outlet of the exchanger.",
        bounds=(0.1, 500), initialize=10
    )
    m.exchanger_cold_side_approach_T = Var(
        m.valid_matches, m.stages,
        doc="Temperature difference between the hot stream outlet and cold "
        "stream inlet of the exchanger.",
        bounds=(0.1, 500), initialize=10
    )
    m.LMTD = Var(
        m.valid_matches, m.stages,
        doc="Log mean temperature difference across the exchanger.",
        bounds=(1, 500), initialize=10
    )
    # Improve LMTD bounds based on T values
    for hot, cold, stg in m.valid_matches * m.stages:
        hot_side_dT_LB = max(0, value(
            m.stage_entry_T[hot, stg].lb - m.stage_exit_T[cold, stg].ub))
        hot_side_dT_UB = max(0, value(
            m.stage_entry_T[hot, stg].ub - m.stage_exit_T[cold, stg].lb))
        cold_side_dT_LB = max(0, value(
            m.stage_exit_T[hot, stg].lb - m.stage_entry_T[cold, stg].ub))
        cold_side_dT_UB = max(0, value(
            m.stage_exit_T[hot, stg].ub - m.stage_entry_T[cold, stg].lb))
        m.LMTD[hot, cold, stg].setlb((
            hot_side_dT_LB * cold_side_dT_LB * (
                hot_side_dT_LB + cold_side_dT_LB) / 2) ** (1 / 3)
        )
        m.LMTD[hot, cold, stg].setub((
            hot_side_dT_UB * cold_side_dT_UB * (
                hot_side_dT_UB + cold_side_dT_UB) / 2) ** (1 / 3)
        )

    m.exchanger_fixed_cost = Var(
        m.stages, m.valid_matches,
        doc="Fixed cost for an exchanger between a hot and cold stream.",
        domain=NonNegativeReals, bounds=(0, 1E5), initialize=0)

    m.exchanger_area = Var(
        m.stages, m.valid_matches,
        doc="Area for an exchanger between a hot and cold stream.",
        domain=NonNegativeReals, bounds=(0, 500), initialize=5)
    m.exchanger_area_cost = Var(
        m.stages, m.valid_matches,
        doc="Capital cost contribution from exchanger area.",
        domain=NonNegativeReals, bounds=(0, 1E5), initialize=1000)

    @m.Constraint(m.hot_process_streams)
    def overall_hot_stream_heat_balance(m, strm):
        return (m.T_in[strm] - m.T_out[strm]) * m.overall_FCp[strm] == (
            sum(m.heat_exchanged[strm, cold, stg]
                for cold in m.cold_streams for stg in m.stages))

    @m.Constraint(m.cold_process_streams)
    def overall_cold_stream_heat_balance(m, strm):
        return (m.T_out[strm] - m.T_in[strm]) * m.overall_FCp[strm] == (
            sum(m.heat_exchanged[hot, strm, stg]
                for hot in m.hot_streams for stg in m.stages))

    @m.Constraint(m.utility_streams)
    def overall_utility_stream_usage(m, strm):
        return m.utility_usage[strm] == (
            sum(m.heat_exchanged[hot, strm, stg]
                for hot in m.hot_process_streams
                for stg in m.stages
                ) if strm in m.cold_utility_streams else 0 +
            sum(m.heat_exchanged[strm, cold, stg]
                for cold in m.cold_process_streams
                for stg in m.stages
                ) if strm in m.hot_utility_streams else 0
        )

    @m.Constraint(m.stages, m.hot_process_streams,
                  doc="Hot side overall heat balance for a stage.")
    def hot_stage_overall_heat_balance(m, stg, strm):
        return ((m.stage_entry_T[strm, stg] - m.stage_exit_T[strm, stg])
                * m.overall_FCp[strm]) == sum(
            m.heat_exchanged[strm, cold, stg]
            for cold in m.cold_streams)

    @m.Constraint(m.stages, m.cold_process_streams,
                  doc="Cold side overall heat balance for a stage.")
    def cold_stage_overall_heat_balance(m, stg, strm):
        return ((m.stage_exit_T[strm, stg] - m.stage_entry_T[strm, stg])
                * m.overall_FCp[strm]) == sum(
            m.heat_exchanged[hot, strm, stg]
            for hot in m.hot_streams)

    @m.Constraint(m.stages, m.hot_process_streams)
    def hot_stream_monotonic_T_decrease(m, stg, strm):
        return m.stage_exit_T[strm, stg] <= m.stage_entry_T[strm, stg]

    @m.Constraint(m.stages, m.cold_process_streams)
    def cold_stream_monotonic_T_increase(m, stg, strm):
        return m.stage_exit_T[strm, stg] >= m.stage_entry_T[strm, stg]

    @m.Constraint(m.stages, m.hot_process_streams)
    def hot_stream_stage_T_link(m, stg, strm):
        return (
            m.stage_exit_T[strm, stg] == m.stage_entry_T[strm, stg + 1]
        ) if stg < num_stages else Constraint.NoConstraint

    @m.Constraint(m.stages, m.cold_process_streams)
    def cold_stream_stage_T_link(m, stg, strm):
        return (
            m.stage_entry_T[strm, stg] == m.stage_exit_T[strm, stg + 1]
        ) if stg < num_stages else Constraint.NoConstraint

    @m.Expression(m.valid_matches, m.stages)
    def exchanger_capacity(m, hot, cold, stg):
        return m.exchanger_area[stg, hot, cold] * (
            m.U[hot, cold] * (
                m.exchanger_hot_side_approach_T[hot, cold, stg] *
                m.exchanger_cold_side_approach_T[hot, cold, stg] *
                (m.exchanger_hot_side_approach_T[hot, cold, stg] +
                 m.exchanger_cold_side_approach_T[hot, cold, stg]) / 2
            ) ** (1 / 3))

    def _exchanger_exists(disj, hot, cold, stg):
        disj.indicator_var.value = 1

        # Log mean temperature difference calculation
        disj.LMTD_calc = Constraint(
            doc="Log mean temperature difference",
            expr=m.LMTD[hot, cold, stg] == (
                m.exchanger_hot_side_approach_T[hot, cold, stg] *
                m.exchanger_cold_side_approach_T[hot, cold, stg] *
                (m.exchanger_hot_side_approach_T[hot, cold, stg] +
                 m.exchanger_cold_side_approach_T[hot, cold, stg]) / 2
            ) ** (1 / 3)
        )
        m.BigM[disj.LMTD_calc] = 160

        # disj.MTD_calc = Constraint(
        #     doc="Mean temperature difference",
        #     expr=m.LMTD[hot, cold, stg] <= (
        #         m.exchanger_hot_side_approach_T[hot, cold, stg] +
        #         m.exchanger_cold_side_approach_T[hot, cold, stg]) / 2
        # )

        # Calculation of the approach temperatures
        if hot in m.hot_utility_streams:
            disj.stage_hot_approach_temperature = Constraint(
                expr=m.exchanger_hot_side_approach_T[hot, cold, stg] <=
                m.T_in[hot] - m.stage_exit_T[cold, stg])
            disj.stage_cold_approach_temperature = Constraint(
                expr=m.exchanger_cold_side_approach_T[hot, cold, stg] <=
                m.T_out[hot] - m.stage_entry_T[cold, stg])
        elif cold in m.cold_utility_streams:
            disj.stage_hot_approach_temperature = Constraint(
                expr=m.exchanger_hot_side_approach_T[hot, cold, stg] <=
                m.stage_entry_T[hot, stg] - m.T_out[cold])
            disj.stage_cold_approach_temperature = Constraint(
                expr=m.exchanger_cold_side_approach_T[hot, cold, stg] <=
                m.stage_exit_T[hot, stg] - m.T_in[cold])
        else:
            disj.stage_hot_approach_temperature = Constraint(
                expr=m.exchanger_hot_side_approach_T[hot, cold, stg] <=
                m.stage_entry_T[hot, stg]
                - m.stage_exit_T[cold, stg])
            disj.stage_cold_approach_temperature = Constraint(
                expr=m.exchanger_cold_side_approach_T[hot, cold, stg] <=
                m.stage_exit_T[hot, stg]
                - m.stage_entry_T[cold, stg])

    def _exchanger_absent(disj, hot, cold, stg):
        disj.indicator_var.value = 0
        disj.no_match_exchanger_cost = Constraint(
            expr=m.exchanger_area_cost[stg, hot, cold] == 0)
        disj.no_match_exchanger_area = Constraint(
            expr=m.exchanger_area[stg, hot, cold] == 0)
        disj.no_match_exchanger_fixed_cost = Constraint(
            expr=m.exchanger_fixed_cost[stg, hot, cold] == 0)
        disj.no_heat_exchange = Constraint(
            expr=m.heat_exchanged[hot, cold, stg] == 0)

    m.exchanger_exists = Disjunct(
        m.valid_matches, m.stages,
        doc="Disjunct for the presence of an exchanger between a "
        "hot stream and a cold stream at a stage.", rule=_exchanger_exists)
    m.exchanger_absent = Disjunct(
        m.valid_matches, m.stages,
        doc="Disjunct for the absence of an exchanger between a "
        "hot stream and a cold stream at a stage.", rule=_exchanger_absent)

    def _exchanger_exists_or_absent(m, hot, cold, stg):
        return [m.exchanger_exists[hot, cold, stg],
                m.exchanger_absent[hot, cold, stg]]
    m.exchanger_exists_or_absent = Disjunction(
        m.valid_matches, m.stages,
        doc="Disjunction between presence or absence of an exchanger between "
        "a hot stream and a cold stream at a stage.",
        rule=_exchanger_exists_or_absent, xor=True)
    # Only hot utility matches in first stage and cold utility matches in last
    # stage
    for hot, cold in m.valid_matches:
        if hot not in m.utility_streams:
            m.exchanger_exists[hot, cold, 1].deactivate()
            m.exchanger_absent[hot, cold, 1].indicator_var.fix(1)
        if cold not in m.utility_streams:
            m.exchanger_exists[hot, cold, num_stages].deactivate()
            m.exchanger_absent[hot, cold, num_stages].indicator_var.fix(1)
    # Exclude utility-stream matches in middle stages
    for hot, cold, stg in m.valid_matches * (m.stages - [1, num_stages]):
        if hot in m.utility_streams or cold in m.utility_streams:
            m.exchanger_exists[hot, cold, stg].deactivate()
            m.exchanger_absent[hot, cold, stg].indicator_var.fix(1)

    @m.Expression(m.utility_streams)
    def utility_cost(m, strm):
        return m.utility_unit_cost[strm] * m.utility_usage[strm]

    m.total_cost = Objective(
        expr=sum(m.utility_cost[strm] for strm in m.utility_streams)
        + sum(m.exchanger_fixed_cost[stg, hot, cold]
              for stg in m.stages
              for hot, cold in m.valid_matches)
        + sum(m.exchanger_area_cost[stg, hot, cold]
              for stg in m.stages
              for hot, cold in m.valid_matches),
        sense=minimize
    )

    return m


def _fix_and_bound(var, val):
    var.fix(val)
    var.setlb(val)
    var.setub(val)



"""Heat integration case study.

This is example 1 of the Yee & Grossmann, 1990 paper "Simultaneous optimization
models for heat integration--II".
DOI: 10.1016/0098-1354(90)85010-8

This is an implementation of the conventional problem.

"""
from __future__ import division

from pyomo.environ import Constraint, log, value

from . import common


def build_conventional(cafaro_approx, num_stages):
    return build_model(cafaro_approx, num_stages)


def build_model(use_cafaro_approximation, num_stages):
    """Build the model."""
    m = common.build_model(use_cafaro_approximation, num_stages)

    for hot, cold, stg in m.valid_matches * m.stages:
        disj = m.exchanger_exists[hot, cold, stg]
        if not use_cafaro_approximation:
            disj.exchanger_area_cost = Constraint(
                expr=m.exchanger_area_cost[stg, hot, cold] * 1E-3 >=
                m.exchanger_area_cost_factor[hot, cold] * 1E-3 *
                m.exchanger_area[stg, hot, cold] ** m.area_cost_exponent)
        else:
            disj.exchanger_area_cost = Constraint(
                expr=m.exchanger_area_cost[stg, hot, cold] * 1E-3 >=
                m.exchanger_area_cost_factor[hot, cold] * 1E-3 * m.cafaro_k
                * log(m.cafaro_b * m.exchanger_area[stg, hot, cold] + 1)
            )
        m.BigM[disj.exchanger_area_cost] = 100

        disj.exchanger_fixed_cost = Constraint(
            expr=m.exchanger_fixed_cost[stg, hot, cold] ==
            m.exchanger_fixed_unit_cost[hot, cold])

        # Area requirement
        disj.exchanger_required_area = Constraint(
            expr=m.exchanger_area[stg, hot, cold] * (
                m.U[hot, cold] * m.LMTD[hot, cold, stg]) >=
            m.heat_exchanged[hot, cold, stg])
        m.BigM[disj.exchanger_required_area] = 5000

    return m



"""Cafaro approximation parameter estimation.

Rather than use the cost relation (1), Cafaro & Grossmann, 2014 (DOI:
10.1016/j.compchemeng.2013.10.001) proposes using (2), which has much better
behaved derivative values near x=0. However, we need to use parameter
estimation in order to derive the correct values of k and b.

This file performs the parameter estimation.

(1) cost = factor * (area) ^ 0.6
(2) cost = factor * k * ln(bx + 1)

"""
from __future__ import division

from pyomo.environ import (ConcreteModel, Constraint, log, NonNegativeReals, SolverFactory, value, Var)


def calculate_cafaro_coefficients(area1, area2, exponent):
    """Calculate the coefficients for the Cafaro approximation.

    Gives the coefficients k and b to approximate a function x^exponent such
    that at the given areas, the following relations apply:

    area1 ^ exponent = k * ln(b * area1 + 1)
    area2 ^ exponent = k * ln(b * area2 + 1)

    Args:
        area1 (float): area to use as the first regression point
        area2 (float): area to use as the second regression point
        exponent (float): exponent to approximate
    """
    m = ConcreteModel()
    m.k = Var(domain=NonNegativeReals)
    m.b = Var(domain=NonNegativeReals)

    m.c1 = Constraint(
        expr=area1 ** exponent == m.k * log(m.b * area1 + 1))
    m.c2 = Constraint(
        expr=area2 ** exponent == m.k * log(m.b * area2 + 1))

    SolverFactory('ipopt').solve(m)

    return value(m.k), value(m.b)



#!/usr/bin/env python
"""
GDPlib open source model library for Generalized Disjunctive Programming
"""
import sys
from logging import warning

from setuptools import setup, find_packages


kwargs = dict(
    name='gdplib',
    packages=find_packages(),
    install_requires=[],
    extras_require={},
    package_data={
        # If any package contains *.template or *.json files, include them:
        '': ['*.template', '*.json']
    },
    scripts=[],
    author='Qi Chen',
    author_email='qichen@andrew.cmu.edu',
    maintainer='Qi Chen',
    url="https://github.com/grossmann-group/gdplib",
    license='BSD 3-clause',
    description="GDPlib open source model library for Generalized Disjunctive Programming",
    long_description=__doc__,
    data_files=[],
    keywords=["pyomo", "generalized disjunctive programming"],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
)

try:
    setup(setup_requires=['setuptools_scm'], use_scm_version=True, **kwargs)
except (ImportError, LookupError):
    default_version = '1.0.0'
    warning('Cannot use .git version: package setuptools_scm not installed '
            'or .git directory not present.')
    print('Defaulting to version: {}'.format(default_version))
    setup(**kwargs)



from os.path import join

from pyomo.common.fileutils import this_file_dir
from pyomo.environ import *
from pyomo.gdp import *

'''Problem from http://www.minlp.org/library/problem/index.php?i=172&lib=GDP
We are minimizing the cost of a design of a plant with parallel processing units and storage tanks
in between. We decide the number and volume of units, and the volume and location of the storage
tanks. The problem is convexified and has a nonlinear objective and global constraints

NOTE: When I refer to 'gams' in the comments, that is Batch101006_BM.gms for now. It's confusing
because the _opt file is different (It has hard-coded bigM parameters so that each constraint 
has the "optimal" bigM).'''


def build_model():

    model = AbstractModel()

    # TODO: it looks like they set a bigM for each j. Which I need to look up how to do...
    model.BigM = Suffix(direction=Suffix.LOCAL)
    model.BigM[None] = 1000


    ## Constants from GAMS
    StorageTankSizeFactor = 2*5 # btw, I know 2*5 is 10... I don't know why it's written this way in GAMS?
    StorageTankSizeFactorByProd = 3
    MinFlow = -log(10000)
    VolumeLB = log(300)
    VolumeUB = log(3500)
    StorageTankSizeLB = log(100)
    StorageTankSizeUB = log(15000)
    UnitsInPhaseUB = log(6)
    UnitsOutOfPhaseUB = log(6)
    # TODO: YOU ARE HERE. YOU HAVEN'T ACTUALLY MADE THESE THE BOUNDS YET, NOR HAVE YOU FIGURED OUT WHOSE
    # BOUNDS THEY ARE. AND THERE ARE MORE IN GAMS.


    ##########
    # Sets
    ##########

    model.PRODUCTS = Set()
    model.STAGES = Set(ordered=True)
    model.PARALLELUNITS = Set(ordered=True)

    # TODO: this seems like an over-complicated way to accomplish this task...
    def filter_out_last(model, j):
        return j != model.STAGES.last()
    model.STAGESExceptLast = Set(initialize=model.STAGES, filter=filter_out_last)


    # TODO: these aren't in the formulation??
    #model.STORAGE_TANKS = Set()


    ###############
    # Parameters
    ###############

    model.HorizonTime = Param()
    model.Alpha1 = Param()
    model.Alpha2 = Param()
    model.Beta1 = Param()
    model.Beta2 = Param()

    model.ProductionAmount = Param(model.PRODUCTS)
    model.ProductSizeFactor = Param(model.PRODUCTS, model.STAGES)
    model.ProcessingTime = Param(model.PRODUCTS, model.STAGES)

    # These are hard-coded in the GAMS file, hence the defaults
    model.StorageTankSizeFactor = Param(model.STAGES, default=StorageTankSizeFactor)
    model.StorageTankSizeFactorByProd = Param(model.PRODUCTS, model.STAGES,
                                              default=StorageTankSizeFactorByProd)

    # TODO: bonmin wasn't happy and I think it might have something to do with this?
    # or maybe issues with convexity or a lack thereof... I don't know yet.
    # I made PRODUCTS ordered so I could do this... Is that bad? And it does index
    # from 1, right?
    def get_log_coeffs(model, k):
        return log(model.PARALLELUNITS.ord(k))

    model.LogCoeffs = Param(model.PARALLELUNITS, initialize=get_log_coeffs)

    # bounds
    model.volumeLB = Param(model.STAGES, default=VolumeLB)
    model.volumeUB = Param(model.STAGES, default=VolumeUB)
    model.storageTankSizeLB = Param(model.STAGES, default=StorageTankSizeLB)
    model.storageTankSizeUB = Param(model.STAGES, default=StorageTankSizeUB)
    model.unitsInPhaseUB = Param(model.STAGES, default=UnitsInPhaseUB)
    model.unitsOutOfPhaseUB = Param(model.STAGES, default=UnitsOutOfPhaseUB)


    ################
    # Variables
    ################

    # TODO: right now these match the formulation. There are more in GAMS...

    # unit size of stage j
    # model.volume = Var(model.STAGES)
    # # TODO: GAMS has a batch size indexed just by products that isn't in the formulation... I'm going
    # # to try to avoid it for the moment...
    # # batch size of product i at stage j
    # model.batchSize = Var(model.PRODUCTS, model.STAGES)
    # # TODO: this is different in GAMS... They index by stages too?
    # # cycle time of product i divided by batch size of product i
    # model.cycleTime = Var(model.PRODUCTS)
    # # number of units in parallel out-of-phase (or in phase) at stage j
    # model.unitsOutOfPhase = Var(model.STAGES)
    # model.unitsInPhase = Var(model.STAGES)
    # # TODO: what are we going to do as a boundary condition here? For that last stage?
    # # size of intermediate storage tank between stage j and j+1
    # model.storageTankSize = Var(model.STAGES)

    # variables for convexified problem
    # TODO: I am beginning to think these are my only variables actually.
    # GAMS never un-logs them, I don't think. And I think the GAMs ones
    # must be the log ones.
    def get_volume_bounds(model, j):
        return (model.volumeLB[j], model.volumeUB[j])
    model.volume_log = Var(model.STAGES, bounds=get_volume_bounds)
    model.batchSize_log = Var(model.PRODUCTS, model.STAGES)
    model.cycleTime_log = Var(model.PRODUCTS)
    def get_unitsOutOfPhase_bounds(model, j):
        return (0, model.unitsOutOfPhaseUB[j])
    model.unitsOutOfPhase_log = Var(model.STAGES, bounds=get_unitsOutOfPhase_bounds)
    def get_unitsInPhase_bounds(model, j):
        return (0, model.unitsInPhaseUB[j])
    model.unitsInPhase_log = Var(model.STAGES, bounds=get_unitsInPhase_bounds)
    def get_storageTankSize_bounds(model, j):
        return (model.storageTankSizeLB[j], model.storageTankSizeUB[j])
    # TODO: these bounds make it infeasible...
    model.storageTankSize_log = Var(model.STAGES, bounds=get_storageTankSize_bounds)

    # binary variables for deciding number of parallel units in and out of phase
    model.outOfPhase = Var(model.STAGES, model.PARALLELUNITS, within=Binary)
    model.inPhase = Var(model.STAGES, model.PARALLELUNITS, within=Binary)

    ###############
    # Objective
    ###############

    def get_cost_rule(model):
        return model.Alpha1 * sum(exp(model.unitsInPhase_log[j] + model.unitsOutOfPhase_log[j] + \
                                              model.Beta1 * model.volume_log[j]) for j in model.STAGES) +\
            model.Alpha2 * sum(exp(model.Beta2 * model.storageTankSize_log[j]) for j in model.STAGESExceptLast)
    model.min_cost = Objective(rule=get_cost_rule)


    ##############
    # Constraints
    ##############

    def processing_capacity_rule(model, j, i):
        return model.volume_log[j] >= log(model.ProductSizeFactor[i, j]) + model.batchSize_log[i, j] - \
            model.unitsInPhase_log[j]
    model.processing_capacity = Constraint(model.STAGES, model.PRODUCTS, rule=processing_capacity_rule)

    def processing_time_rule(model, j, i):
        return model.cycleTime_log[i] >= log(model.ProcessingTime[i, j]) - model.batchSize_log[i, j] - \
            model.unitsOutOfPhase_log[j]
    model.processing_time = Constraint(model.STAGES, model.PRODUCTS, rule=processing_time_rule)

    def finish_in_time_rule(model):
        return model.HorizonTime >= sum(model.ProductionAmount[i]*exp(model.cycleTime_log[i]) \
                                        for i in model.PRODUCTS)
    model.finish_in_time = Constraint(rule=finish_in_time_rule)


    ###############
    # Disjunctions
    ###############

    def storage_tank_selection_disjunct_rule(disjunct, selectStorageTank, j):
        model = disjunct.model()
        def volume_stage_j_rule(disjunct, i):
            return model.storageTankSize_log[j] >= log(model.StorageTankSizeFactor[j]) + \
                model.batchSize_log[i, j]
        def volume_stage_jPlus1_rule(disjunct, i):
            return model.storageTankSize_log[j] >= log(model.StorageTankSizeFactor[j]) + \
                model.batchSize_log[i, j+1]
        def batch_size_rule(disjunct, i):
            return inequality(-log(model.StorageTankSizeFactorByProd[i,j]),
                              model.batchSize_log[i,j] - model.batchSize_log[i, j+1],
                              log(model.StorageTankSizeFactorByProd[i,j]))
        def no_batch_rule(disjunct, i):
            return model.batchSize_log[i,j] - model.batchSize_log[i,j+1] == 0

        if selectStorageTank:
            disjunct.volume_stage_j = Constraint(model.PRODUCTS, rule=volume_stage_j_rule)
            disjunct.volume_stage_jPlus1 = Constraint(model.PRODUCTS,
                                                      rule=volume_stage_jPlus1_rule)
            disjunct.batch_size = Constraint(model.PRODUCTS, rule=batch_size_rule)
        else:
            # The formulation says 0, but GAMS has this constant.
            # 04/04: Francisco says volume should be free:
            # disjunct.no_volume = Constraint(expr=model.storageTankSize_log[j] == MinFlow)
            disjunct.no_batch = Constraint(model.PRODUCTS, rule=no_batch_rule)
    model.storage_tank_selection_disjunct = Disjunct([0,1], model.STAGESExceptLast,
                                           rule=storage_tank_selection_disjunct_rule)

    def select_storage_tanks_rule(model, j):
        return [model.storage_tank_selection_disjunct[selectTank, j] for selectTank in [0,1]]
    model.select_storage_tanks = Disjunction(model.STAGESExceptLast, rule=select_storage_tanks_rule)

    # though this is a disjunction in the GAMs model, it is more efficiently formulated this way:
    # TODO: what on earth is k?
    def units_out_of_phase_rule(model, j):
        return model.unitsOutOfPhase_log[j] == sum(model.LogCoeffs[k] * model.outOfPhase[j,k] \
                                                   for k in model.PARALLELUNITS)
    model.units_out_of_phase = Constraint(model.STAGES, rule=units_out_of_phase_rule)

    def units_in_phase_rule(model, j):
        return model.unitsInPhase_log[j] == sum(model.LogCoeffs[k] * model.inPhase[j,k] \
                                                for k in model.PARALLELUNITS)
    model.units_in_phase = Constraint(model.STAGES, rule=units_in_phase_rule)

    # and since I didn't do the disjunction as a disjunction, we need the XORs:
    def units_out_of_phase_xor_rule(model, j):
        return sum(model.outOfPhase[j,k] for k in model.PARALLELUNITS) == 1
    model.units_out_of_phase_xor = Constraint(model.STAGES, rule=units_out_of_phase_xor_rule)

    def units_in_phase_xor_rule(model, j):
        return sum(model.inPhase[j,k] for k in model.PARALLELUNITS) == 1
    model.units_in_phase_xor = Constraint(model.STAGES, rule=units_in_phase_xor_rule)

    return model.create_instance(join(this_file_dir(), 'batch_processing.dat'))


if __name__ == "__main__":
    m = build_model()
    TransformationFactory('gdp.bigm').apply_to(m)
    SolverFactory('gams').solve(m, solver='baron', tee=True, add_options=['option optcr=1e-6;'])
    m.min_cost.display()



"""Heat integration case study.

This is example 1 of the Yee & Grossmann, 1990 paper "Simultaneous optimization
models for heat integration--II".
DOI: 10.1016/0098-1354(90)85010-8

This is a modification to support the incorporation of standardized exchanger
modules.

This is a further modification to support discretization of area simplifying
the nonlinear expressions.

"""
from __future__ import division

from pyomo.environ import (Binary, Constraint, log, Set, Var)
from pyomo.gdp import Disjunct, Disjunction

from . import common


def build_require_modular(cafaro_approx, num_stages):
    m = build_model(cafaro_approx, num_stages)
    # Require modular
    for hot, cold, stg in m.valid_matches * m.stages:
        disj = m.exchanger_exists[hot, cold, stg]
        disj.modular.indicator_var.fix(1)
        disj.conventional.deactivate()

    for hot, cold in m.valid_matches:
        lowest_price = float('inf')
        for size in sorted(m.possible_sizes, reverse=True):
            current_size_cost = (m.modular_size_cost[hot, cold, size] +
                                 m.modular_fixed_cost[hot, cold, size])
            if current_size_cost > lowest_price:
                for stg in m.stages:
                    m.module_size_active[hot, cold, stg, size].fix(0)
            else:
                lowest_price = current_size_cost

    return m


def build_modular_option(cafaro_approx, num_stages):
    m = build_model(cafaro_approx, num_stages)

    for hot, cold in m.valid_matches:
        lowest_price = float('inf')
        for size in sorted(m.possible_sizes, reverse=True):
            current_size_cost = (m.modular_size_cost[hot, cold, size] +
                                 m.modular_fixed_cost[hot, cold, size])
            if current_size_cost > lowest_price:
                for stg in m.stages:
                    m.module_size_active[hot, cold, stg, size].fix(0)
            else:
                lowest_price = current_size_cost

    return m


def build_model(use_cafaro_approximation, num_stages):
    """Build the model."""
    m = common.build_model(use_cafaro_approximation, num_stages)

    m.possible_sizes = Set(initialize=[10 * (i + 1) for i in range(50)])
    m.module_size_active = Var(
        m.valid_matches, m.stages, m.possible_sizes,
        doc="Total area of modular exchangers for each match.",
        domain=Binary, initialize=0)

    num_modules_required = {}
    for size in m.possible_sizes:
        # For each possible size, calculate the number of exchangers of each
        # area required to satisfy that size.
        # Use progressively smaller exchangers to satisfy the size
        remaining_size = size
        module_sizes = sorted(m.module_sizes, reverse=True)
        for area in module_sizes:
            num_modules_required[size, area] = remaining_size // area
            remaining_size = remaining_size % area

    @m.Param(m.valid_matches, m.possible_sizes, m.module_sizes,
             doc="Number of exchangers of each area required to "
             "yield a certain total size.")
    def modular_num_exchangers(m, hot, cold, size, area):
        return num_modules_required[size, area]

    @m.Param(m.valid_matches, m.possible_sizes,
             doc="Area cost for each modular exchanger size.")
    def modular_size_cost(m, hot, cold, size):
        return sum(m.modular_num_exchangers[hot, cold, size, area] *
                   m.module_area_cost[hot, cold, area]
                   for area in m.module_sizes)

    @m.Param(m.valid_matches, m.possible_sizes,
             doc="Fixed cost for each modular exchanger size.")
    def modular_fixed_cost(m, hot, cold, size):
        return sum(m.modular_num_exchangers[hot, cold, size, area] *
                   m.module_fixed_unit_cost
                   for area in m.module_sizes)

    m.LMTD_discretize = Var(
        m.hot_streams, m.cold_streams, m.stages, m.possible_sizes,
        doc="Discretized log mean temperature difference",
        bounds=(0, 500), initialize=0
    )

    for hot, cold, stg in m.valid_matches * m.stages:
        disj = m.exchanger_exists[hot, cold, stg]
        disj.conventional = Disjunct()
        if not use_cafaro_approximation:
            disj.conventional.exchanger_area_cost = Constraint(
                expr=m.exchanger_area_cost[stg, hot, cold] * 1E-3 >=
                m.exchanger_area_cost_factor[hot, cold] * 1E-3 *
                m.exchanger_area[stg, hot, cold] ** m.area_cost_exponent)
        else:
            disj.conventional.exchanger_area_cost = Constraint(
                expr=m.exchanger_area_cost[stg, hot, cold] * 1E-3 >=
                m.exchanger_area_cost_factor[hot, cold] * 1E-3 * m.cafaro_k
                * log(m.cafaro_b * m.exchanger_area[stg, hot, cold] + 1)
            )
        m.BigM[disj.conventional.exchanger_area_cost] = 100

        disj.conventional.exchanger_fixed_cost = Constraint(
            expr=m.exchanger_fixed_cost[stg, hot, cold] ==
            m.exchanger_fixed_unit_cost[hot, cold])

        @disj.conventional.Constraint(m.possible_sizes)
        def no_modules(_, size):
            return m.module_size_active[hot, cold, stg, size] == 0

        # Area requirement
        disj.conventional.exchanger_required_area = Constraint(
            expr=m.exchanger_area[stg, hot, cold] *
            m.U[hot, cold] * m.LMTD[hot, cold, stg] >=
            m.heat_exchanged[hot, cold, stg])
        m.BigM[disj.conventional.exchanger_required_area] = 5000

        disj.modular = Disjunct()
        disj.modular.choose_one_config = Constraint(
            expr=sum(
                m.module_size_active[hot, cold, stg, size]
                for size in m.possible_sizes) == 1
        )
        disj.modular.exchanger_area_cost = Constraint(
            expr=m.exchanger_area_cost[stg, hot, cold] * 1E-3 ==
            sum(m.modular_size_cost[hot, cold, size] * 1E-3 *
                m.module_size_active[hot, cold, stg, size]
                for size in m.possible_sizes)
        )
        disj.modular.exchanger_fixed_cost = Constraint(
            expr=m.exchanger_fixed_cost[stg, hot, cold] ==
            sum(m.modular_fixed_cost[hot, cold, size] * 1E-3 *
                m.module_size_active[hot, cold, stg, size]
                for size in m.possible_sizes))

        disj.modular.discretize_area = Constraint(
            expr=m.exchanger_area[stg, hot, cold] == sum(
                area * m.module_size_active[hot, cold, stg, area]
                for area in m.possible_sizes)
        )

        disj.modular.discretized_LMTD = Constraint(
            expr=m.LMTD[hot, cold, stg] == sum(
                m.LMTD_discretize[hot, cold, stg, size]
                for size in m.possible_sizes
            )
        )

        @disj.modular.Constraint(m.possible_sizes)
        def discretized_LMTD_LB(disj, size):
            return (
                m.LMTD[hot, cold, stg].lb
                * m.module_size_active[hot, cold, stg, size]
            ) <= m.LMTD_discretize[hot, cold, stg, size]

        @disj.modular.Constraint(m.possible_sizes)
        def discretized_LMTD_UB(disj, size):
            return m.LMTD_discretize[hot, cold, stg, size] <= (
                m.LMTD[hot, cold, stg].ub
                * m.module_size_active[hot, cold, stg, size]
            )

        disj.modular.exchanger_required_area = Constraint(
            expr=m.U[hot, cold] * sum(
                area * m.LMTD_discretize[hot, cold, stg, area]
                for area in m.possible_sizes) >=
            m.heat_exchanged[hot, cold, stg])

        disj.modular_or_not = Disjunction(
            expr=[disj.modular, disj.conventional])

    return m



from __future__ import division

import os

import pandas as pd
from pyomo.environ import (
    ConcreteModel, Constraint, Integers, maximize, Objective, Param, RangeSet, SolverFactory, summation,
    TransformationFactory, value, Var, )
from pyomo.gdp import Disjunct, Disjunction


def build_model(case="Growth"):
    m = ConcreteModel()
    m.months = RangeSet(0, 120, doc="10 years")
    m.discount_rate = Param(initialize=0.08, doc="8%")
    m.conv_setup_time = Param(initialize=12)
    m.modular_setup_time = Param(initialize=3)
    m.modular_teardown_time = Param(initialize=3)
    m.conventional_salvage_value = Param(initialize=0.05, doc="5%")
    m.module_salvage_value = Param(initialize=0.3, doc="30%")

    @m.Param(m.months, doc="Discount factor")
    def discount_factor(m, mo):
        return (1 + m.discount_rate / 12) ** (-mo / 12)

    @m.Param(m.months, doc="Value of each unit of production")
    def production_value(m, mo):
        return (7 / 12) * m.discount_factor[mo]

    xls_data = pd.read_excel(
        os.path.join(os.path.dirname(__file__), "market_size.xlsx"), sheet_name="single_market", index_col=0)

    @m.Param(m.months)
    def market_demand(m, mo):
        return float(xls_data[case][mo])

    m.conv_size = Var(bounds=(25, 350), initialize=25,
                      doc="Size of conventional plant")
    m.conv_base_cost = Param(initialize=1000, doc="Cost for size 25")
    m.conv_exponent = Param(initialize=0.6)
    m.conv_cost = Var(bounds=(0, value(m.conv_base_cost * (m.conv_size.ub / 25) ** m.conv_exponent)))
    m.module_base_cost = Param(initialize=1000, doc="Cost for size 25")

    m.production = Var(m.months, bounds=(0, 350))
    m.num_modules = Var(m.months, domain=Integers, bounds=(0, 35))
    m.modules_purchased = Var(m.months, domain=Integers, bounds=(0, 35))
    m.modules_sold = Var(m.months, domain=Integers, bounds=(0, 35))

    @m.Param(m.months)
    def module_unit_cost(m, mo):
        return m.module_base_cost * m.discount_factor[mo]

    @m.Constraint(m.months)
    def demand_limit(m, mo):
        return m.production[mo] <= m.market_demand[mo]

    @m.Expression(m.months)
    def revenue(m, mo):
        return m.production_value[mo] * m.production[mo]

    @m.Expression()
    def conv_salvage_val(m):
        return m.conv_cost * m.conventional_salvage_value * m.discount_factor[120]

    @m.Expression(m.months)
    def module_buy_cost(m, mo):
        return m.module_base_cost * m.discount_factor[mo] * m.modules_purchased[mo]

    @m.Expression(m.months)
    def module_sell_value(m, mo):
        return m.module_base_cost * m.discount_factor[mo] * m.module_salvage_value * m.modules_sold[mo]

    @m.Expression()
    def module_final_salvage(m):
        mo = max(m.months)
        return m.module_base_cost * m.discount_factor[mo] * m.module_salvage_value * m.num_modules[mo]

    m.profit = Objective(
        expr=sum(m.revenue[:])
        - m.conv_cost
        + m.conv_salvage_val
        - summation(m.module_buy_cost)
        + summation(m.module_sell_value)
        + m.module_final_salvage,
        sense=maximize)

    _build_conventional_disjunct(m)
    _build_modular_disjunct(m)
    m.conventional_or_modular = Disjunction(expr=[m.conventional, m.modular])

    return m


def _build_conventional_disjunct(m):
    m.conventional = Disjunct()

    m.conventional.cost_calc = Constraint(
        expr=m.conv_cost == (
            m.conv_base_cost * (m.conv_size / 25) ** m.conv_exponent))

    @m.conventional.Constraint(m.months)
    def conv_production_limit(conv_disj, mo):
        if mo < m.conv_setup_time:
            return m.production[mo] == 0
        else:
            return m.production[mo] <= m.conv_size

    @m.conventional.Constraint()
    def no_modules(conv_disj):
        return sum(m.num_modules[:]) + sum(m.modules_purchased[:]) + sum(m.modules_sold[:]) == 0


def _build_modular_disjunct(m):
    m.modular = Disjunct()

    @m.modular.Constraint(m.months)
    def module_balance(disj, mo):
        existing_modules = 0 if mo == 0 else m.num_modules[mo - 1]
        new_modules = 0 if mo < m.modular_setup_time else m.modules_purchased[mo - m.modular_setup_time]
        sold_modules = m.modules_sold[mo]
        return m.num_modules[mo] == existing_modules + new_modules - sold_modules

    @m.modular.Constraint(m.months)
    def modular_production_limit(mod_disj, mo):
        return m.production[mo] <= 25 * m.num_modules[mo]


def display_conventional(m, writer, sheet_name):
    df = pd.DataFrame(
        list([
            mo,
            m.production[mo].value,
            m.market_demand[mo],
            m.conv_size.value if mo >= m.conv_setup_time else 0
        ] for mo in m.months),
        columns=("Month", "Production", "Demand", "Capacity")
    ).set_index('Month').round(2)
    df.to_excel(writer, sheet_name)
    print('Conventional Profit', round(value(m.profit)))
    print('Conventional Revenue', round(value(sum(m.revenue[:]))))
    print('Conventional Size', round(value(m.conv_size)))
    print('Conventional Build Cost', round(value(m.conv_cost)))
    print('Conventional Salvage Value', round(value(m.conv_salvage_val)))
    print()


def display_modular(m, writer, sheet_name):
    df = pd.DataFrame(
        list([
            mo,
            m.production[mo].value,
            m.market_demand[mo],
            m.num_modules[mo].value * 25,
            m.num_modules[mo].value,
            m.modules_purchased[mo].value,
            m.modules_sold[mo].value] for mo in m.months
        ),
        columns=("Month", "Production", "Demand", "Capacity",
                 "Num Modules", "Add Modules", "Sold Modules")
    ).set_index('Month').round(2)
    df.to_excel(writer, sheet_name)
    print('Modular Profit', round(value(m.profit)))
    print('Modular Revenue', round(value(sum(m.revenue[:]))))
    print('Modular Revenue before conventional startup', round(value(sum(m.revenue[mo] for mo in m.months if mo < 12))))
    print('Modular Build Cost', round(value(sum(m.module_buy_cost[:]))))
    print('Modules Purchased', round(value(sum(m.modules_purchased[:]))))
    print('Modular Nondiscount Cost', round(value(m.module_base_cost * sum(m.modules_purchased[:]))))
    print('Modular Sale Credit', round(value(sum(m.module_sell_value[:]))))
    print('Modular Final Salvage Credit', round(value(m.module_final_salvage)))
    print()


if __name__ == "__main__":
    cases = ['Growth', 'Dip', 'Decay']
    conv_size_vals = {}
    with pd.ExcelWriter('cap_expand_config.xlsx') as writer:
        for case in cases:
            print(case)
            m = build_model(case)
            m.conventional.indicator_var.fix(1)
            m.modular.deactivate()
            TransformationFactory('gdp.bigm').apply_to(m, bigM=7000)
            SolverFactory('gams').solve(m, solver='baron')
            conv_size_vals[case] = value(m.conv_size)
            display_conventional(m, writer, 'conv_%s' % case)
            del m

            m = build_model(case)
            m.modular.indicator_var.fix(1)
            m.conventional.deactivate()
            TransformationFactory('gdp.chull').apply_to(m)
            SolverFactory('gurobi').solve(m)
            display_modular(m, writer, 'mod_%s' % case)

    # Run conventional case with fixed values
